import sys
import argparse
import copy
import warnings

from cogdl.datasets import try_adding_dataset_args
from cogdl.models import try_adding_model_args
from cogdl.wrappers import fetch_data_wrapper, fetch_model_wrapper
from cogdl.utils import build_args_from_dict
from cogdl.wrappers.default_match import get_wrappers_name


def get_parser():
    parser = argparse.ArgumentParser(conflict_handler="resolve")
    # fmt: off
    parser.add_argument("--seed", default=[1], type=int, nargs="+", metavar="N",
                        help="pseudo random number generator seed")
    parser.add_argument("--epochs", default=500, type=int)
    parser.add_argument("--max-epoch", default=None, type=int)
    parser.add_argument("--patience", type=int, default=100)
    parser.add_argument("--lr", default=0.01, type=float)
    parser.add_argument("--weight-decay", default=0, type=float)
    parser.add_argument("--n-warmup-steps", type=int, default=0)
    parser.add_argument("--split", default=[0], type=int, nargs="+", metavar="N")

    parser.add_argument("--checkpoint-path", type=str, default="./checkpoints/model.pt", help="path to save model")
    parser.add_argument("--save-emb-path", type=str, default=None, help="path to save embeddings")
    parser.add_argument("--load-emb-path", type=str, default=None, help="path to load embeddings")
    parser.add_argument("--resume-training", action="store_true")
    parser.add_argument("--logger", type=str, default=None)
    parser.add_argument("--log-path", type=str, default=".", help="path to save logs")
    parser.add_argument("--project", type=str, default="cogdl-exp", help="project name for wandb")

    parser.add_argument("--use-best-config", action="store_true", help="use best config")
    parser.add_argument("--unsup", action="store_true")
    parser.add_argument("--nstage", type=int, default=1)
    parser.add_argument("--eval-step", type=int, default=1)
    parser.add_argument("--n-trials", type=int, default=3)

    parser.add_argument("--devices", default=[0], type=int, nargs="+", help="which GPU to use")
    parser.add_argument("--cpu", action="store_true", help="use CPU instead of CUDA")
    parser.add_argument("--cpu-inference", action="store_true", help="do validation and test in cpu")
    parser.add_argument("--distributed", action="store_true")
    parser.add_argument("--progress-bar", type=str, default="epoch")
    parser.add_argument("--local_rank", type=int, default=0)
    parser.add_argument("--master-port", type=int, default=13425)
    parser.add_argument("--master-addr", type=str, default="localhost")

    parser.add_argument("--no-test", action="store_true")

    parser.add_argument("--actnn", action="store_true")
    parser.add_argument("--rp-ratio", type=int, default=1)
    # fmt: on
    return parser


def add_data_wrapper_args(parser):
    group = parser.add_argument_group("Data wrapper configuration")
    # fmt: off
    group.add_argument("--dw", "-t", type=str, default=None, metavar="DWRAPPER", required=False,
                       help="Data Wrapper")
    # fmt: on
    return group


def add_model_wrapper_args(parser):
    group = parser.add_argument_group("Trainer configuration")
    # fmt: off
    group.add_argument("--mw", type=str, default=None, metavar="MWRAPPER", required=False,
                       help="Model Wrapper")
    # fmt: on
    return group


def add_dataset_args(parser):
    group = parser.add_argument_group("Dataset and data loading")
    # fmt: off
    group.add_argument("--dataset", "-dt", metavar="DATASET", nargs="+", required=True,
                       help="Dataset")
    # fmt: on
    return group


def add_model_args(parser):
    group = parser.add_argument_group("Model configuration")
    # fmt: off
    group.add_argument("--model", "-m", default="autognn", metavar="MODEL", nargs="+", required=False,
                       help="Model Architecture")
    # fmt: on
    return group


def get_training_parser():
    parser = get_parser()
    add_dataset_args(parser)
    add_model_args(parser)
    add_data_wrapper_args(parser)
    add_model_wrapper_args(parser)
    return parser


def get_display_data_parser():
    parser = get_parser()
    add_dataset_args(parser)
    parser.add_argument("--depth", default=3, type=int)

    return parser


def get_download_data_parser():
    parser = get_parser()
    add_dataset_args(parser)

    return parser


def get_default_args(dataset, model, **kwargs):
    if not isinstance(dataset, list):
        dataset = [dataset]
    if not isinstance(model, list):
        model = [model]
    sys.argv = [sys.argv[0], "-m"] + model + ["-dt"] + dataset
    if "mw" in kwargs and kwargs["mw"] is not None:
        sys.argv += ["--mw"] + [kwargs["mw"]]
    if "dw" in kwargs and kwargs["dw"] is not None:
        sys.argv += ["--dw"] + [kwargs["dw"]]

    # The parser doesn"t know about specific args, so we parse twice.
    parser = get_training_parser()
    args, _ = parser.parse_known_args()
    args = parse_args_and_arch(parser, args)
    for key, value in kwargs.items():
        args.__setattr__(key, value)
    return args


def get_diff_args(args1, args2):
    d1 = copy.deepcopy(args1.__dict__)
    d2 = args2.__dict__
    for k in d2.keys():
        d1.pop(k, None)
    return build_args_from_dict(d1)


def parse_args_and_arch(parser, args):
    # Add *-specific args to parser.
    for model in args.model:
        try_adding_model_args(model, parser)

    for dataset in args.dataset:
        try_adding_dataset_args(dataset, parser)

    if len(args.model) > 1:
        warnings.warn("Please ensure that models could use the same model wrapper!")
    default_wrappers = get_wrappers_name(args.model[0])
    if default_wrappers is not None:
        mw, dw = default_wrappers
    else:
        mw, dw = None, None

    if args.dw is not None:
        dw = args.dw
    if dw is None:
        # warnings.warn("Using default data wrapper ('node_classification_dw') for training!")
        dw = "node_classification_dw"
    if hasattr(fetch_data_wrapper(dw), "add_args"):
        fetch_data_wrapper(dw).add_args(parser)

    if args.mw is not None:
        mw = args.mw
    if mw is None:
        # warnings.warn("Using default model wrapper ('node_classification_mw') for training!")
        mw = "node_classification_mw"
    if hasattr(fetch_model_wrapper(mw), "add_args"):
        fetch_model_wrapper(mw).add_args(parser)

    # Parse a second time.
    args = parser.parse_args()
    args.mw = mw
    args.dw = dw
    return args



from .compgcn import CompGCN
from .dgi import DGIModel
from .disengcn import DisenGCN
from .gcn import GCN
from .gcnii import GCNII
from .gdc_gcn import GDC_GCN
from .grace import GRACE
from .graphsage import Graphsage
from .mvgrl import MVGRL
from .patchy_san import PatchySAN
from .ppnp import PPNP
from .rgcn import RGCN
from .sgc import sgc
from .revgcn import RevGCN, RevGEN, RevGAT
from .deepergcn import DeeperGCN, ResGNNLayer

__all__ = [
    "CompGCN",
    "DGIModel",
    "DisenGCN",
    "GCN",
    "GCNII",
    "GDC_GCN",
    "GRACE",
    "Graphsage",
    "MVGRL",
    "PatchySAN",
    "PPNP",
    "RGCN",
    "sgc",
    "RevGCN",
    "RevGAT",
    "RevGEN",
    "DeeperGCN",
    "ResGNNLayer",
]



import torch
import torch.nn.functional as F
from cogdl.layers import GCNLayer

from .. import BaseModel
from .gcn import GCN


class GAE(GCN):
    @classmethod
    def build_model_from_args(cls, args):
        return cls(args.num_features, args.hidden_size, args.num_layers, args.dropout)

    def __init__(self, in_feats, hidden_size, num_layers, dropout):
        super(GAE, self).__init__(in_feats, hidden_size, 1, num_layers, dropout)

    def make_loss(self, data, adj):
        embeddings = self.embed(data)
        return (
            F.binary_cross_entropy(F.softmax(torch.mm(embeddings, embeddings.t())), adj, reduction="sum")
            / data.x.shape[0]
        )

    def get_features(self, data):
        return self.embed(data).detach()


class VGAE(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--hidden-size", type=int, default=64)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(args.num_features, args.hidden_size)

    def __init__(self, num_features, hidden_size):
        super(VGAE, self).__init__()
        self.num_features = num_features
        self.hidden_size = hidden_size
        self.conv1 = GCNLayer(self.num_features, self.hidden_size)
        self.conv2_mean = GCNLayer(self.hidden_size, self.hidden_size)
        self.conv2_var = GCNLayer(self.hidden_size, self.hidden_size)

    def reparameterize(self, mean, log_var):
        sigma = torch.exp(log_var)
        z = mean + torch.randn_like(log_var) * sigma
        return z

    def encode(self, graph):
        graph.add_remaining_self_loops()
        graph.sym_norm()

        h = graph.x
        h = self.conv1(graph, h)
        h = F.relu(h)
        mean = self.conv2_mean(graph, h)
        log_var = self.conv2_var(graph, h)
        return mean, log_var

    def decode(self, x):
        return torch.sigmoid(torch.matmul(x, x.t()))

    def forward(self, graph):
        mean, log_var = self.encode(graph)
        return self.reparameterize(mean, log_var)

    def get_features(self, graph):
        return self.forward(graph).detach()

    def make_loss(self, data, adj):
        mean, log_var = self.encode(data)
        z = self.reparameterize(mean, log_var)
        mat = self.decode(z)
        recon_loss = F.binary_cross_entropy(mat, adj, reduction="sum")
        var = torch.exp(log_var)
        kl_loss = 0.5 * torch.mean(torch.sum(mean * mean + var - log_var - 1, dim=1))
        print("recon_loss = %.3f, kl_loss = %.3f" % (recon_loss, kl_loss))
        return recon_loss + kl_loss



import importlib

from .base_model_wrapper import ModelWrapper, EmbeddingModelWrapper


def register_model_wrapper(name):
    """
    New data wrapper types can be added to cogdl with the :func:`register_model_wrapper`
    function decorator.

    Args:
        name (str): the name of the model_wrapper
    """

    def register_model_wrapper_cls(cls):
        print("The `register_model_wrapper` API is deprecated!")
        return cls

    return register_model_wrapper_cls


def fetch_model_wrapper(name):
    if isinstance(name, type):
        return name
    if name in SUPPORTED_MW:
        path = ".".join(SUPPORTED_MW[name].split(".")[:-1])
        module = importlib.import_module(path)
    else:
        raise NotImplementedError(f"Failed to import {name} ModelWrapper.")
    class_name = SUPPORTED_MW[name].split(".")[-1]
    return getattr(module, class_name)


SUPPORTED_MW = {
    "dgi_mw": "cogdl.wrappers.model_wrapper.node_classification.DGIModelWrapper",
    "gcnmix_mw": "cogdl.wrappers.model_wrapper.node_classification.GCNMixModelWrapper",
    "grace_mw": "cogdl.wrappers.model_wrapper.node_classification.GRACEModelWrapper",
    "grand_mw": "cogdl.wrappers.model_wrapper.node_classification.GrandModelWrapper",
    "mvgrl_mw": "cogdl.wrappers.model_wrapper.node_classification.MVGRLModelWrapper",
    "self_auxiliary_mw": "cogdl.wrappers.model_wrapper.node_classification.SelfAuxiliaryModelWrapper",
    "graphsage_mw": "cogdl.wrappers.model_wrapper.node_classification.GraphSAGEModelWrapper",
    "unsup_graphsage_mw": "cogdl.wrappers.model_wrapper.node_classification.UnsupGraphSAGEModelWrapper",
    "m3s_mw": "cogdl.wrappers.model_wrapper.node_classification.M3SModelWrapper",
    "network_embedding_mw": "cogdl.wrappers.model_wrapper.node_classification.NetworkEmbeddingModelWrapper",
    "node_classification_mw": "cogdl.wrappers.model_wrapper.node_classification.NodeClfModelWrapper",
    "correct_smooth_mw": "cogdl.wrappers.model_wrapper.node_classification.CorrectSmoothModelWrapper",
    "pprgo_mw": "cogdl.wrappers.model_wrapper.node_classification.PPRGoModelWrapper",
    "sagn_mw": "cogdl.wrappers.model_wrapper.node_classification.SAGNModelWrapper",
    "gcc_mw": "cogdl.wrappers.model_wrapper.pretraining.GCCModelWrapper",
    "embedding_link_prediction_mw": "cogdl.wrappers.model_wrapper.link_prediction.EmbeddingLinkPredictionModelWrapper",
    "gnn_kg_link_prediction_mw": "cogdl.wrappers.model_wrapper.link_prediction.GNNKGLinkPredictionModelWrapper",
    "gnn_link_prediction_mw": "cogdl.wrappers.model_wrapper.link_prediction.GNNLinkPredictionModelWrapper",
    "heterogeneous_embedding_mw": "cogdl.wrappers.model_wrapper.heterogeneous.HeterogeneousEmbeddingModelWrapper",
    "heterogeneous_gnn_mw": "cogdl.wrappers.model_wrapper.heterogeneous.HeterogeneousGNNModelWrapper",
    "multiplex_embedding_mw": "cogdl.wrappers.model_wrapper.heterogeneous.MultiplexEmbeddingModelWrapper",
    "graph_classification_mw": "cogdl.wrappers.model_wrapper.graph_classification.GraphClassificationModelWrapper",
    "graph_embedding_mw": "cogdl.wrappers.model_wrapper.graph_classification.GraphEmbeddingModelWrapper",
    "infograph_mw": "cogdl.wrappers.model_wrapper.graph_classification.InfoGraphModelWrapper",
    "agc_mw": "cogdl.wrappers.model_wrapper.clustering.AGCModelWrapper",
    "daegc_mw": "cogdl.wrappers.model_wrapper.clustering.DAEGCModelWrapper",
    "gae_mw": "cogdl.wrappers.model_wrapper.clustering.GAEModelWrapper",
}



import numba
import numpy as np
import torch
import scipy.sparse as sp
import random

# from cogdl.utils.rwalk import random_walk as c_random_walk


@numba.njit(cache=True, parallel=True)
def random_walk(start, length, indptr, indices, p=0.0):
    """
    Parameters:
        start : np.array(dtype=np.int32)
        length : int
        indptr : np.array(dtype=np.int32)
        indices : np.array(dtype=np.int32)
        p : float
    Return:
        list(np.array(dtype=np.int32))
    """
    result = [np.zeros(length, dtype=np.int32)] * len(start)
    for i in numba.prange(len(start)):
        result[i] = _random_walk(start[i], length, indptr, indices, p)
    return result


@numba.njit(cache=True)
def _random_walk(node, length, indptr, indices, p=0.0):
    result = [numba.int32(0)] * length
    result[0] = numba.int32(node)
    i = numba.int32(1)
    _node = node
    while i < length:
        start = indptr[node]
        end = indptr[node + 1]
        sample = random.randint(start, end - 1)
        node = indices[sample]
        if np.random.uniform(0, 1) > p:
            result[i] = node
        else:
            result[i] = _node
        i += 1
    return np.array(result, dtype=np.int32)


class RandomWalker(object):
    def __init__(self, adj=None, num_nodes=None):
        if adj is None:
            self.indptr = None
            self.indices = None
        else:
            if isinstance(adj, torch.Tensor):
                if num_nodes is None:
                    num_nodes = int(torch.max(adj)) + 1
                row, col = adj.cpu().numpy()
                data = np.ones(row.shape[0])
                adj = sp.csr_matrix((data, (row, col)), shape=(num_nodes, num_nodes))
            adj = adj.tocsr()

            self.indptr = adj.indptr
            self.indices = adj.indices

    def build_up(self, adj, num_nodes):
        if self.indptr is not None:
            return
        if isinstance(adj, torch.Tensor) or isinstance(adj, tuple):
            row, col = adj
            if num_nodes is None:
                num_nodes = int(max(row.max(), col.max())) + 1
            row, col = row.cpu().numpy(), col.cpu().numpy()
            data = np.ones(row.shape[0])
            adj = sp.csr_matrix((data, (row, col)), shape=(num_nodes, num_nodes))
        adj = adj.tocsr()

        self.indptr = adj.indptr
        self.indices = adj.indices

    def walk(self, start, walk_length, restart_p=0.0):
        assert self.indptr is not None, "Please build the adj_list first"
        if isinstance(start, torch.Tensor):
            start = start.cpu().numpy()
        if isinstance(start, list):
            start = np.asarray(start, dtype=np.int32)
        result = random_walk(start, walk_length, self.indptr, self.indices, restart_p)
        result = np.array(result, dtype=np.int64)
        return result



import torch
import torch.nn.functional as F

from .. import ModelWrapper
from cogdl.wrappers.tools.wrapper_utils import evaluate_clustering


class GAEModelWrapper(ModelWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--num-clusters", type=int, default=7)
        parser.add_argument("--cluster-method", type=str, default="kmeans", help="option: kmeans or spectral")
        parser.add_argument("--evaluation", type=str, default="full", help="option: full or NMI")
        # fmt: on

    def __init__(self, model, optimizer_cfg, num_clusters, cluster_method="kmeans", evaluation="full"):
        super(GAEModelWrapper, self).__init__()
        self.model = model
        self.num_clusters = num_clusters
        self.optimizer_cfg = optimizer_cfg
        self.cluster_method = cluster_method
        self.full = evaluation == "full"

    def train_step(self, subgraph):
        graph = subgraph
        loss = self.model.make_loss(graph, graph.adj_mx)
        return loss

    def test_step(self, subgraph):
        graph = subgraph
        features_matrix = self.model(graph)
        features_matrix = features_matrix.detach().cpu().numpy()
        return evaluate_clustering(
            features_matrix, graph.y, self.cluster_method, self.num_clusters, graph.num_nodes, self.full
        )

    def pre_stage(self, stage, data_w):
        if stage == 0:
            data = data_w.get_dataset().data
            adj_mx = torch.sparse_coo_tensor(
                torch.stack(data.edge_index),
                torch.ones(data.edge_index[0].shape[0]),
                torch.Size([data.x.shape[0], data.x.shape[0]]),
            ).to_dense()
            data.adj_mx = adj_mx

    def setup_optimizer(self):
        lr, wd = self.optimizer_cfg["lr"], self.optimizer_cfg["weight_decay"]
        return torch.optim.Adam(self.parameters(), lr=lr, weight_decay=wd)



import argparse
import numpy as np
import torch
from sklearn.metrics import auc, f1_score, precision_recall_curve, roc_auc_score

from cogdl.data import Graph
from .. import EmbeddingModelWrapper


def get_score(embs, node1, node2):
    vector1 = embs[int(node1)]
    vector2 = embs[int(node2)]
    return np.dot(vector1, vector2) / (np.linalg.norm(vector1) * np.linalg.norm(vector2))


def evaluate(embs, true_edges, false_edges):
    true_list = list()
    prediction_list = list()
    for edge in true_edges:
        true_list.append(1)
        prediction_list.append(get_score(embs, edge[0], edge[1]))

    for edge in false_edges:
        true_list.append(0)
        prediction_list.append(get_score(embs, edge[0], edge[1]))

    sorted_pred = prediction_list[:]
    sorted_pred.sort()
    threshold = sorted_pred[-len(true_edges)]

    y_pred = np.zeros(len(prediction_list), dtype=np.int32)
    for i in range(len(prediction_list)):
        if prediction_list[i] >= threshold:
            y_pred[i] = 1

    y_true = np.array(true_list)
    y_scores = np.array(prediction_list)
    ps, rs, _ = precision_recall_curve(y_true, y_scores)
    return roc_auc_score(y_true, y_scores), f1_score(y_true, y_pred), auc(rs, ps)


def evaluate_multiplex(all_embs, test_data):
    total_roc_auc, total_f1_score, total_pr_auc = [], [], []
    for key in test_data.keys():
        embs = all_embs[key]
        roc_auc, f1_score, pr_auc = evaluate(embs, test_data[key][0], test_data[key][1])
        total_roc_auc.append(roc_auc)
        total_f1_score.append(f1_score)
        total_pr_auc.append(pr_auc)
    assert len(total_roc_auc) > 0
    roc_auc, f1_score, pr_auc = (
        np.mean(total_roc_auc),
        np.mean(total_f1_score),
        np.mean(total_pr_auc),
    )
    print(f"Test ROC-AUC = {roc_auc:.4f}, F1 = {f1_score:.4f}, PR-AUC = {pr_auc:.4f}")
    return dict(ROC_AUC=roc_auc, PR_AUC=pr_auc, F1=f1_score)


class MultiplexEmbeddingModelWrapper(EmbeddingModelWrapper):
    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        """Add task-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--hidden-size", type=int, default=200)
        parser.add_argument("--eval-type", type=str, default='all', nargs='+')
        # fmt: on

    def __init__(self, model, hidden_size=200, eval_type="all"):
        super(MultiplexEmbeddingModelWrapper, self).__init__()
        self.model = model
        self.hidden_size = hidden_size
        self.eval_type = eval_type

    def train_step(self, batch):
        if hasattr(self.model, "multiplicity"):
            all_embs = self.model(batch)
        else:
            all_embs = dict()
            for key in batch.keys():
                if self.eval_type == "all" or key in self.eval_type:
                    G = Graph(edge_index=torch.LongTensor(batch[key]).transpose(0, 1))
                    embs = self.model(G, return_dict=True)
                all_embs[key] = embs
        return all_embs

    def test_step(self, batch):
        all_embs, test_data = batch
        return evaluate_multiplex(all_embs, test_data)



import random
import time

import networkx as nx
import numpy as np
from gensim.models import Word2Vec

from cogdl.utils import alias_draw, alias_setup
from .. import BaseModel


class Node2vec(BaseModel):
    r"""The node2vec model from the `"node2vec: Scalable feature learning for networks"
    <http://dl.acm.org/citation.cfm?doid=2939672.2939754>`_ paper

    Args:
        hidden_size (int) : The dimension of node representation.
        walk_length (int) : The walk length.
        walk_num (int) : The number of walks to sample for each node.
        window_size (int) : The actual context size which is considered in language model.
        worker (int) : The number of workers for word2vec.
        iteration (int) : The number of training iteration in word2vec.
        p (float) : Parameter in node2vec.
        q (float) : Parameter in node2vec.
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument('--walk-length', type=int, default=80,
                            help='Length of walk per source. Default is 80.')
        parser.add_argument('--walk-num', type=int, default=40,
                            help='Number of walks per source. Default is 40.')
        parser.add_argument('--window-size', type=int, default=5,
                            help='Window size of skip-gram model. Default is 5.')
        parser.add_argument('--worker', type=int, default=10,
                            help='Number of parallel workers. Default is 10.')
        parser.add_argument('--iteration', type=int, default=10,
                            help='Number of iterations. Default is 10.')
        parser.add_argument('--p', type=float, default=1.0,
                            help='Parameter in node2vec. Default is 1.0.')
        parser.add_argument('--q', type=float, default=1.0,
                            help='Parameter in node2vec. Default is 1.0.')
        parser.add_argument("--hidden-size", type=int, default=128)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.hidden_size,
            args.walk_length,
            args.walk_num,
            args.window_size,
            args.worker,
            args.iteration,
            args.p,
            args.q,
        )

    def __init__(self, dimension, walk_length, walk_num, window_size, worker, iteration, p, q):
        super(Node2vec, self).__init__()
        self.dimension = dimension
        self.walk_length = walk_length
        self.walk_num = walk_num
        self.window_size = window_size
        self.worker = worker
        self.iteration = iteration
        self.p = p
        self.q = q

    def forward(self, graph, return_dict=False):
        G = graph.to_networkx()
        self.G = G
        is_directed = nx.is_directed(self.G)
        for i, j in G.edges():
            G[i][j]["weight"] = G[i][j].get("weight", 1.0)
            if not is_directed:
                G[j][i]["weight"] = G[j][i].get("weight", 1.0)
        self._preprocess_transition_probs()
        walks = self._simulate_walks(self.walk_num, self.walk_length)
        walks = [[str(node) for node in walk] for walk in walks]
        model = Word2Vec(
            walks,
            size=self.dimension,
            window=self.window_size,
            min_count=0,
            sg=1,
            workers=self.worker,
            iter=self.iteration,
        )
        id2node = dict([(vid, node) for vid, node in enumerate(G.nodes())])
        embeddings = np.asarray([model.wv[str(id2node[i])] for i in range(len(id2node))])

        if return_dict:
            features_matrix = dict()
            for vid, node in enumerate(G.nodes()):
                features_matrix[node] = embeddings[vid]
        else:
            features_matrix = np.zeros((graph.num_nodes, embeddings.shape[1]))
            nx_nodes = G.nodes()
            features_matrix[nx_nodes] = embeddings[np.arange(graph.num_nodes)]
        return features_matrix

    def _node2vec_walk(self, walk_length, start_node):
        # Simulate a random walk starting from start node.
        G = self.G
        alias_nodes = self.alias_nodes
        alias_edges = self.alias_edges

        walk = [start_node]

        while len(walk) < walk_length:
            cur = walk[-1]
            cur_nbrs = list(G.neighbors(cur))
            if len(cur_nbrs) > 0:
                if len(walk) == 1:
                    walk.append(cur_nbrs[alias_draw(alias_nodes[cur][0], alias_nodes[cur][1])])
                else:
                    prev = walk[-2]
                    next = cur_nbrs[alias_draw(alias_edges[(prev, cur)][0], alias_edges[(prev, cur)][1])]
                    walk.append(next)
            else:
                break

        return walk

    def _simulate_walks(self, num_walks, walk_length):
        # Repeatedly simulate random walks from each node.
        G = self.G
        walks = []
        nodes = list(G.nodes())
        print("Walk iteration:")
        for walk_iter in range(num_walks):
            if walk_iter % 10 == 0:
                print(str(walk_iter + 1), "/", str(num_walks))
            random.shuffle(nodes)
            for node in nodes:
                walks.append(self._node2vec_walk(walk_length=walk_length, start_node=node))

        return walks

    def _get_alias_edge(self, src, dst):
        # Get the alias edge setup lists for a given edge.
        G = self.G
        unnormalized_probs = []
        for dst_nbr in G.neighbors(dst):
            if dst_nbr == src:
                unnormalized_probs.append(G[dst][dst_nbr]["weight"] / self.p)
            elif G.has_edge(dst_nbr, src):
                unnormalized_probs.append(G[dst][dst_nbr]["weight"])
            else:
                unnormalized_probs.append(G[dst][dst_nbr]["weight"] / self.q)
        norm_const = sum(unnormalized_probs)
        normalized_probs = [float(u_prob) / norm_const for u_prob in unnormalized_probs]

        return alias_setup(normalized_probs)

    def _preprocess_transition_probs(self):
        # Preprocessing of transition probabilities for guiding the random walks.
        G = self.G
        is_directed = nx.is_directed(self.G)

        print(len(list(G.nodes())))
        print(len(list(G.edges())))

        s = time.time()
        alias_nodes = {}
        for node in G.nodes():
            unnormalized_probs = [G[node][nbr]["weight"] for nbr in G.neighbors(node)]
            norm_const = sum(unnormalized_probs)
            normalized_probs = [float(u_prob) / norm_const for u_prob in unnormalized_probs]
            alias_nodes[node] = alias_setup(normalized_probs)

        t = time.time()
        print("alias_nodes", t - s)

        alias_edges = {}
        s = time.time()

        if is_directed:
            for edge in G.edges():
                alias_edges[edge] = self._get_alias_edge(edge[0], edge[1])
        else:
            for edge in G.edges():
                alias_edges[edge] = self._get_alias_edge(edge[0], edge[1])
                alias_edges[(edge[1], edge[0])] = self._get_alias_edge(edge[1], edge[0])

        t = time.time()
        print("alias_edges", t - s)

        self.alias_nodes = alias_nodes
        self.alias_edges = alias_edges

        return



import math

import torch
import torch.nn as nn

from cogdl.utils import get_activation


class LinearLayer(nn.Module):
    def __init__(self, in_features, out_features, bias=True):
        super(LinearLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.kaiming_uniform_(self.weight, mode="fan_out", a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / nn.math.sqrt(fan_in)
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input):
        return torch.nn.functional.linear(input, self.weight, self.bias)


class PPRGoLayer(nn.Module):
    def __init__(self, in_feats, hidden_size, out_feats, num_layers, dropout, activation="relu"):
        super(PPRGoLayer, self).__init__()
        self.dropout = dropout
        self.nlayers = num_layers
        shapes = [hidden_size] * (num_layers - 1) + [out_feats]
        self.layers = nn.ModuleList()
        self.layers.append(LinearLayer(in_feats, hidden_size, bias=False))
        for i in range(num_layers - 1):
            self.layers.append(nn.Linear(shapes[i], shapes[i + 1], bias=False))
        self.activation = get_activation(activation)

    def forward(self, x):
        h = x
        for i, layer in enumerate(self.layers):
            h = nn.functional.dropout(h, p=self.dropout, training=self.training)
            h = layer(h)
            if i != self.nlayers - 1:
                h = self.activation(h)
        return h



import torch

# import torch.fft
import torch.nn as nn
import torch.nn.functional as F

from cogdl.data import Graph
from cogdl.utils import row_normalization
from cogdl.utils.link_prediction_utils import GNNLinkPredict, sampling_edge_uniform, cal_mrr
from .. import BaseModel


# def ccorr(a, b):
#     return torch.fft.irfft(
#         torch.multiply(torch.fft.rfft(a, dim=1).conj(), torch.fft.rfft(b, dim=1)), n=a.shape[-1], dim=1
#     )


class BasesRelEmbLayer(nn.Module):
    def __init__(self, num_bases, num_rels, in_feats):
        super(BasesRelEmbLayer, self).__init__()
        self.num_bases = num_bases
        self.num_resl = num_rels
        self.in_feats = in_feats
        self.weight = nn.Parameter(torch.Tensor(num_bases, in_feats))
        self.alpha = nn.Parameter(torch.Tensor(2 * num_rels, num_bases))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.weight)
        nn.init.xavier_uniform_(self.alpha)

    def forward(self):
        weight = torch.matmul(self.alpha, self.weight)
        return weight


class CompGCNLayer(nn.Module):
    """
    Implementation of CompGCN in paper `"Composition-based Multi-Relational Graph Convolutional Networks"`
    <https://arxiv.org/abs/1911.03082>

    Parameters
    ----------
    in_feats : int
        Size of each input embedding
    out_feats : int
        Size of each output embedding
    num_rel : int
        The number of relations
    opn : str
        Operation to mix two head and tail embeddings
    num_bases : Optional[Int]
        The number of basis, default : ``None``.
    activation : Function
    dropout : float
    bias : bool
    """

    def __init__(
        self, in_feats, out_feats, num_rels, opn="mult", num_bases=None, activation=lambda x: x, dropout=0.0, bias=True
    ):
        super(CompGCNLayer, self).__init__()
        self.in_feats = in_feats
        self.out_feats = out_feats
        self.num_rels = num_rels
        self.opn = opn
        self.use_bases = num_bases is not None and num_bases > 0

        self.weight_in = self.get_param(in_feats, out_feats)
        self.weight_out = self.get_param(in_feats, out_feats)
        self.weight_rel = self.get_param(in_feats, out_feats)
        self.weight_loop = self.get_param(in_feats, out_feats)
        self.loop_rel = self.get_param(1, in_feats)

        if self.use_bases:
            self.basis_weight = BasesRelEmbLayer(num_bases, num_rels, in_feats)
        else:
            self.register_buffer("basis_weight", None)

        self.dropout = dropout
        self.activation = activation
        if bias:
            self.bias = nn.Parameter(torch.zeros(out_feats))
        else:
            self.register_buffer("bias", None)
        self.bn = nn.BatchNorm1d(out_feats)

    def get_param(self, num_in, num_out):
        weight = nn.Parameter(torch.Tensor(num_in, num_out))
        nn.init.xavier_normal_(weight.data)
        return weight

    def forward(self, graph, x, rel_embed):
        device = x.device
        if self.use_bases:
            rel_embed = self.basis_weight()
        edge_index = graph.edge_index
        edge_type = graph.edge_attr
        rel_embed = torch.cat((rel_embed, self.loop_rel), dim=0)
        num_edges = edge_index[0].shape[0] // 2
        num_entities = x.shape[0]

        row, col = edge_index
        i_row, i_col = row[:num_edges], col[:num_edges]
        rev_row, rev_col = row[num_edges:], col[num_edges:]
        loop_index = torch.stack((torch.arange(num_entities), torch.arange(num_entities))).to(device)
        types, rev_types = edge_type[:num_edges], edge_type[num_edges:]
        loop_types = torch.full((num_entities,), rel_embed.shape[0] - 1, dtype=torch.long).to(device)

        in_norm = row_normalization(num_entities, i_row, i_col)
        rev_norm = row_normalization(num_entities, rev_row, rev_col)

        emb = self.message_passing(x, rel_embed, (i_row, i_col), types, "in", in_norm)
        rev_emb = self.message_passing(x, rel_embed, (rev_row, rev_col), rev_types, "out", rev_norm)
        loop_emb = self.message_passing(x, rel_embed, loop_index, loop_types, "loop")

        out = 1 / 3 * (emb + rev_emb + loop_emb)
        if self.bias is not None:
            out += self.bias
        out = self.bn(out)
        return self.activation(out), torch.matmul(rel_embed, self.weight_rel)[:-1]

    def message_passing(self, x, rel_embed, edge_index, edge_types, mode, edge_weight=None):
        device = x.device
        tail_emb = x[edge_index[1]]
        rel_emb = rel_embed[edge_types]
        weight = getattr(self, f"weight_{mode}")

        trans_embed = self.rel_transform(tail_emb, rel_emb)
        trans_embed = torch.matmul(trans_embed, weight)
        dim = trans_embed.shape[1]
        if edge_weight is not None:
            trans_embed = trans_embed * edge_weight.unsqueeze(-1)
        embed = (
            torch.zeros(x.shape[0], dim)
            .to(device)
            .scatter_add_(0, edge_index[0].unsqueeze(-1).repeat(1, dim), trans_embed)
        )
        return F.dropout(embed, p=self.dropout, training=self.training)

    def rel_transform(self, ent_embed, rel_embed):
        # if self.opn == "corr":
        #     trans_embed = ccorr(ent_embed, rel_embed)
        if self.opn == "sub":
            trans_embed = ent_embed - rel_embed
        elif self.opn == "mult":
            trans_embed = ent_embed * rel_embed
        else:
            raise NotImplementedError(f"{self.opn}")
        return trans_embed


class CompGCN(nn.Module):
    def __init__(
        self, num_entities, num_rels, num_bases, in_feats, hidden_size, out_feats, layers, dropout, activation, opn
    ):
        super(CompGCN, self).__init__()
        self.opn = opn
        self.num_rels = num_rels
        self.num_entities = num_entities
        if num_bases is not None and num_bases > 0:
            self.init_rel = nn.Embedding(num_bases, in_feats)
        else:
            self.init_rel = nn.Embedding(2 * num_rels, in_feats)

        self.convs = nn.ModuleList()
        if num_bases > 0:
            self.convs.append(
                CompGCNLayer(
                    in_feats=in_feats,
                    out_feats=hidden_size,
                    num_rels=num_rels,
                    opn=self.opn,
                    num_bases=num_bases,
                    activation=activation,
                    dropout=dropout,
                )
            )
        else:
            self.convs.append(
                CompGCNLayer(
                    in_feats=in_feats,
                    out_feats=hidden_size,
                    num_rels=num_rels,
                    opn=self.opn,
                    activation=activation,
                    dropout=dropout,
                )
            )
        if layers == 2:
            self.convs.append(
                CompGCNLayer(
                    in_feats=hidden_size,
                    out_feats=out_feats,
                    num_rels=num_rels,
                    opn=self.opn,
                    activation=activation,
                    dropout=dropout,
                )
            )

    def forward(self, graph, x):
        rel_embed = self.init_rel.weight
        node_embed = x
        for layer in self.convs:
            node_embed, rel_embed = layer(graph, node_embed, rel_embed)
        return node_embed, rel_embed


class LinkPredictCompGCN(GNNLinkPredict, BaseModel):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--hidden-size", type=int, default=200)
        parser.add_argument("--penalty", type=float, default=0.001)
        parser.add_argument("--dropout", type=float, default=0.3)
        parser.add_argument("--num-bases", type=int, default=10)
        parser.add_argument("--num-layers", type=int, default=1)
        parser.add_argument("--sampling-rate", type=float, default=0.01)
        parser.add_argument("--lbl_smooth", type=float, default=0.1)
        parser.add_argument("--opn", type=str, default="sub")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            num_entities=args.num_entities,
            num_rels=args.num_rels,
            hidden_size=args.hidden_size,
            num_bases=args.num_bases,
            sampling_rate=args.sampling_rate,
            penalty=args.penalty,
            layers=args.num_layers,
            dropout=args.dropout,
            lbl_smooth=args.lbl_smooth,
            opn=args.opn,
        )

    def __init__(
        self,
        num_entities,
        num_rels,
        hidden_size,
        num_bases=0,
        layers=1,
        sampling_rate=0.01,
        penalty=0.001,
        dropout=0.0,
        lbl_smooth=0.1,
        opn="sub",
    ):
        BaseModel.__init__(self)
        GNNLinkPredict.__init__(self)
        activation = F.tanh
        self.model = CompGCN(
            num_entities,
            num_rels,
            num_bases,
            hidden_size // 2,
            hidden_size,
            hidden_size,
            layers,
            dropout,
            activation,
            opn,
        )
        self.emb = nn.Embedding(num_entities, hidden_size // 2)
        self.sampling_rate = sampling_rate
        self.penalty = penalty
        self.num_rels = num_rels
        self.num_entities = num_entities
        self.cache_index = None
        self.lbl_smooth = lbl_smooth

    def add_reverse_edges(self, edge_index, edge_types):
        row, col = edge_index
        _row = torch.cat([row, col])
        _col = torch.cat([col, row])
        edge_index = torch.cat((_row, _col), dim=1)
        edge_types_rev = edge_types + self.num_rels
        edge_types = torch.cat([edge_types, edge_types_rev])
        return edge_index, edge_types

    def forward(self, graph):
        edge_index = torch.stack(graph.edge_index)
        # edge_index, edge_types = self.add_reverse_edges(edge_index, edge_types)
        reindexed_node, reindexed_edge_index = torch.unique(edge_index, return_inverse=True, sorted=True)
        self.cache_index = reindexed_node
        node_embed = self.emb(reindexed_node)
        with graph.local_graph():
            graph.edge_index = reindexed_edge_index
            node_embed, rel_embed = self.model(graph, node_embed)
        return node_embed, rel_embed

    def loss(self, data: Graph, scoring):
        row, col = data.edge_index
        edge_types = data.edge_attr
        edge_index = torch.stack([row, col])

        self.get_edge_set(edge_index, edge_types)
        batch_edges, batch_attr, samples, rels, labels = sampling_edge_uniform(
            (row, col),
            edge_types,
            self.edge_set,
            self.sampling_rate,
            self.num_rels,
            label_smoothing=self.lbl_smooth,
            num_entities=self.num_entities,
        )
        with data.local_graph():
            data.edge_index = batch_edges
            data.edge_attr = batch_attr
            node_embed, rel_embed = self.forward(data)

        sampled_nodes, reindexed_edges = torch.unique(samples, sorted=True, return_inverse=True)
        assert (self.cache_index == sampled_nodes).any()
        loss_n = self._loss(
            node_embed[reindexed_edges[0]], node_embed[reindexed_edges[1]], rel_embed[rels], labels, scoring
        )
        loss_r = self.penalty * self._regularization([self.emb(sampled_nodes), rel_embed])
        return loss_n + loss_r

    def predict(self, graph):
        device = next(self.parameters()).device
        indices = torch.arange(0, self.num_entities).to(device)
        x = self.emb(indices)
        node_embed, rel_embed = self.model(graph, x)
        return node_embed, rel_embed



import os
import numpy as np
import torch
from torch.utils.cpp_extension import load

path = os.path.join(os.path.dirname(__file__))

# SPMM


try:
    spmm = load(
        name="spmm",
        extra_ldflags=["-lcusparse"],
        sources=[os.path.join(path, "spmm/spmm.cpp"), os.path.join(path, "spmm/spmm_kernel.cu")],
        verbose=False,
    )
    sddmm = load(
        name="sddmm",
        sources=[os.path.join(path, "spmm/sddmm.cpp"), os.path.join(path, "spmm/sddmm_kernel.cu")],
        verbose=False,
    )

    def csrspmm(rowptr, colind, x, csr_data, sym=False, actnn=False):
        if actnn:
            return ActSPMMFunction.apply(rowptr, colind, x, csr_data, sym)
        return SPMMFunction.apply(rowptr, colind, x, csr_data, sym)


except Exception:
    csrspmm = None


try:
    spmm_cpu = load(
        name="spmm_cpu",
        extra_cflags=["-fopenmp"],
        sources=[os.path.join(path, "spmm/spmm_cpu.cpp")],
        verbose=False,
    )
    spmm_cpu = spmm_cpu.csr_spmm_cpu
except Exception as e:
    print(e)
    spmm_cpu = None


class SPMMFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, rowptr, colind, feat, edge_weight_csr=None, sym=False):
        if edge_weight_csr is None:
            out = spmm.csr_spmm_no_edge_value(rowptr, colind, feat)
        else:
            out = spmm.csr_spmm(rowptr, colind, edge_weight_csr, feat)
        if edge_weight_csr is not None and edge_weight_csr.requires_grad:
            ctx.backward_csc = (rowptr, colind, feat, edge_weight_csr, sym)
        else:
            ctx.backward_csc = (rowptr, colind, edge_weight_csr, sym)
        return out

    @staticmethod
    def backward(ctx, grad_out):
        if len(ctx.backward_csc) == 5:
            rowptr, colind, feat, edge_weight_csr, sym = ctx.backward_csc
        else:
            rowptr, colind, edge_weight_csr, sym = ctx.backward_csc
        if edge_weight_csr is not None:
            grad_out = grad_out.contiguous()
            if sym:
                colptr, rowind, edge_weight_csc = rowptr, colind, edge_weight_csr
            else:
                colptr, rowind, edge_weight_csc = spmm.csr2csc(rowptr, colind, edge_weight_csr)
            grad_feat = spmm.csr_spmm(colptr, rowind, edge_weight_csc, grad_out)
            if edge_weight_csr.requires_grad:
                grad_edge_weight = sddmm.csr_sddmm(rowptr, colind, grad_out, feat)
            else:
                grad_edge_weight = None
        else:
            if sym is False:
                colptr, rowind, edge_weight_csc = spmm.csr2csc(rowptr, colind, edge_weight_csr)
                grad_feat = spmm.csr_spmm_no_edge_value(colptr, rowind, grad_out)
            else:
                grad_feat = spmm.csr_spmm_no_edge_value(rowptr, colind, grad_out)
            grad_edge_weight = None
        return None, None, grad_feat, grad_edge_weight, None


try:
    from actnn.ops import quantize_activation, dequantize_activation
except Exception:
    pass


class ActSPMMFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, rowptr, colind, feat, edge_weight_csr=None, sym=False):
        if edge_weight_csr is None:
            out = spmm.csr_spmm_no_edge_value(rowptr, colind, feat)
        else:
            out = spmm.csr_spmm(rowptr, colind, edge_weight_csr, feat)
        if edge_weight_csr is not None and edge_weight_csr.requires_grad:
            quantized = quantize_activation(feat, None)
            ctx.backward_csc = (rowptr, colind, quantized, edge_weight_csr, sym)
            ctx.other_args = feat.shape
        else:
            ctx.backward_csc = (rowptr, colind, edge_weight_csr, sym)
        return out

    @staticmethod
    def backward(ctx, grad_out):
        if len(ctx.backward_csc) == 5:
            rowptr, colind, quantized, edge_weight_csr, sym = ctx.backward_csc
            q_input_shape = ctx.other_args
            feat = dequantize_activation(quantized, q_input_shape)
            del quantized
        else:
            rowptr, colind, edge_weight_csr, sym = ctx.backward_csc
        del ctx.backward_csc

        if edge_weight_csr is not None:
            grad_out = grad_out.contiguous()
            if sym:
                colptr, rowind, edge_weight_csc = rowptr, colind, edge_weight_csr
            else:
                colptr, rowind, edge_weight_csc = spmm.csr2csc(rowptr, colind, edge_weight_csr)
            grad_feat = spmm.csr_spmm(colptr, rowind, edge_weight_csc, grad_out)
            if edge_weight_csr.requires_grad:
                grad_edge_weight = sddmm.csr_sddmm(rowptr, colind, grad_out, feat)
            else:
                grad_edge_weight = None
        else:
            if sym is False:
                colptr, rowind, edge_weight_csc = spmm.csr2csc(rowptr, colind, edge_weight_csr)
                grad_feat = spmm.csr_spmm_no_edge_value(colptr, rowind, grad_out)
            else:
                grad_feat = spmm.csr_spmm_no_edge_value(rowptr, colind, grad_out)
            grad_edge_weight = None
        return None, None, grad_feat, grad_edge_weight, None



import math
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from .. import ModelWrapper
from cogdl.models.nn.mlp import MLP
from cogdl.data import DataLoader
from cogdl.wrappers.tools.wrapper_utils import evaluate_graph_embeddings_using_svm


class InfoGraphModelWrapper(ModelWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--sup", action="store_true")
        # fmt: on

    def __init__(self, model, optimizer_cfg, sup=False):
        super(InfoGraphModelWrapper, self).__init__()
        self.model = model
        hidden_size = optimizer_cfg["hidden_size"]
        model_num_layers = model.num_layers
        self.local_dis = FF(model_num_layers * hidden_size, hidden_size)
        self.global_dis = FF(model_num_layers * hidden_size, hidden_size)

        self.optimizer_cfg = optimizer_cfg
        self.sup = sup
        self.criterion = torch.nn.MSELoss()

    def train_step(self, batch):
        if self.sup:
            pred = self.model.sup_forward(batch, batch.x)
            loss = self.sup_loss(pred, batch)
        else:
            graph_feat, node_feat = self.model.unsup_forward(batch, batch.x)
            loss = self.unsup_loss(graph_feat, node_feat, batch.batch)
        return loss

    def test_step(self, dataset):
        device = self.device
        dataloader = DataLoader(dataset, batch_size=32, shuffle=False)
        preds = []
        with torch.no_grad():
            for batch in dataloader:
                preds.append(self.model(batch.to(device)))
        preds = torch.cat(preds).cpu().numpy()
        labels = np.array([g.y.item() for g in dataset])
        result = evaluate_graph_embeddings_using_svm(preds, labels)

        self.note("test_metric", result["acc"])
        self.note("std", result["std"])

    def setup_optimizer(self):
        cfg = self.optimizer_cfg
        return torch.optim.Adam(
            [
                {"params": self.model.parameters()},
                {"params": self.global_dis.parameters()},
                {"params": self.local_dis.parameters()},
            ],
            lr=cfg["lr"],
            weight_decay=cfg["weight_decay"],
        )

    def sup_loss(self, pred, batch):
        pred = F.softmax(pred, dim=1)
        loss = self.criterion(pred, batch)
        loss += self.unsup_loss(batch.x, batch.edge_index, batch.batch)[1]
        return loss

    def unsup_loss(self, graph_feat, node_feat, batch):
        local_encode = self.local_dis(node_feat)
        global_encode = self.global_dis(graph_feat)

        num_graphs = graph_feat.shape[0]
        num_nodes = node_feat.shape[0]

        pos_mask = torch.zeros((num_nodes, num_graphs)).to(batch.device)
        neg_mask = torch.ones((num_nodes, num_graphs)).to(batch.device)
        for nid, gid in enumerate(batch):
            pos_mask[nid][gid] = 1
            neg_mask[nid][gid] = 0
        glob_local_mi = torch.mm(local_encode, global_encode.t())
        loss = self.mi_loss(pos_mask, neg_mask, glob_local_mi, num_nodes, num_nodes * (num_graphs - 1))
        return loss

    @staticmethod
    def mi_loss(pos_mask, neg_mask, mi, pos_div, neg_div):
        pos_mi = pos_mask * mi
        neg_mi = neg_mask * mi

        pos_loss = (-math.log(2.0) + F.softplus(-pos_mi)).sum()
        neg_loss = (-math.log(2.0) + F.softplus(-neg_mi) + neg_mi).sum()
        # pos_loss = F.softplus(-pos_mi).sum()
        # neg_loss = (F.softplus(neg_mi)).sum()
        # pos_loss = pos_mi.sum()
        # neg_loss = neg_mi.sum()
        return pos_loss / pos_div + neg_loss / neg_div


class FF(nn.Module):
    r"""Residual MLP layers.

    ..math::
        out = \mathbf{MLP}(x) + \mathbf{Linear}(x)

    Paramaters
    ----------
    in_feats : int
        Size of each input sample
    out_feats : int
        Size of each output sample
    """

    def __init__(self, in_feats, out_feats):
        super(FF, self).__init__()
        self.block = MLP(in_feats, out_feats, out_feats, num_layers=3)
        self.shortcut = nn.Linear(in_feats, out_feats)

    def forward(self, x):
        return F.relu(self.block(x)) + self.shortcut(x)



BEST_CONFIGS = {
    "chebyshev": {"general": {}},
    "dropedge_gcn": {"general": {}},
    "gat": {
        "general": {"lr": 0.005, "epochs": 1000},
        "citeseer": {"weight_decay": 0.001},
        "pubmed": {"weight_decay": 0.001},
        "ppi-large": {
            "nhead": 4,
            "hidden_size": 256,
            "lr": 0.005,
            "residual": True,
            "num_layers": 3,
            "weight_decay": 0.0,
            "attn_drop": 0,
            "dropout": 0,
        },
    },
    "gcn": {
        "general": {},
        "ppi-large": {
            "lr": 0.005,
            "hidden_size": 2048,
            "num_layers": 2,
            "dropout": 0.3,
            "epochs": 7000,
            # 78.04
        },
        "flickr": {
            "lr": 0.01,
            "dropout": 0.5,
            "weight_decay": 0.0,
            "hidden_size": 256,
        },
    },
    "sage": {
        "general": {},
    },
    "gcnii": {
        "general": {
            "epochs": 1000,
            "dropout": 0.5,
            "wd1": 0.001,
            "wd2": 5e-4,
        },
        "cora": {
            "num_layers": 64,
            "hidden_size": 64,
            "dropout": 0.6,
        },
        "citeseer": {
            "num_layers": 32,
            "hidden_size": 256,
            "lr": 0.001,
            "patience": 200,
            "epochs": 2000,
            "lmbda": 0.6,
            "dropout": 0.7,
        },
        "pubmed": {
            "num_layers": 16,
            "hidden_size": 256,
            "lmbda": 0.4,
            "dropout": 0.5,
            "wd1": 5e-4,
        },
        "reddit": {
            "num_layers": 3,
            "hidden_size": 256,
            "alpha": 0.2,
            "wd1": 0,
            "wd2": 0,
            "lr": 0.001,
            "residual": True,
            "epochs": 1000,
            "lmbda": 1.0,
            # 96.42
        },
        "flickr": {
            "lr": 0.002,
            "num_layers": 4,
            "residual": True,
            "wd1": 0.0005,
            "wd2": 0.0005,
            "lmbda": 1.0,
            "alpha": 0.1,
        }
        # 52.54
    },
    "gdc_gcn": {
        "general": {"hidden_size": 16},
    },
    "grand": {
        "general": {
            "epochs": 1000,
        },
        "cora": {
            "order": 8,
            "sample": 4,
            "lam": 1.0,
            "tem": 0.5,
            "alpha": 0.5,
            "patience": 200,
            "input_dropout": 0.5,
            "hidden_dropout": 0.5,
        },
        "citeseer": {
            "order": 2,
            "sample": 2,
            "lam": 0.7,
            "tem": 0.3,
            "alpha": 0.5,
            "input_dropout": 0.0,
            "hidden_dropout": 0.2,
            "patience": 200,
        },
        "pubmed": {
            "order": 5,
            "sample": 4,
            "lam": 1.0,
            "tem": 0.2,
            "alpha": 0.5,
            "lr": 0.2,
            "bn": True,
            "input_dropout": 0.6,
            "hidden_dropout": 0.8,
        },
    },
    "graphsage": {
        "general": {},
    },
    "sgc": {
        "general": {
            "hidden_size": 16,
            "dropout": 0.5,
        },
    },
    "sgcpn": {
        "general": {
            "lr": 0.005,
            "epochs": 1000,
            "patience": 1000,
            "norm_mode": "PN",
            "norm_scale": 10,
            "dropout": 0.6,
        },
    },
    "sign": {
        "general": {
            "lr": 0.00005,
            "hidden_size": 2048,
            "dropout": 0.5,
            "dropedge_rate": 0.2,
        },
    },
    "srgcn": {
        "general": {
            "lr": 0.005,
            "epochs": 1000,
        },
        "cora": {"dropout": 0.6},
        "citeseer": {"dropout": 0.6},
    },
    "unet": {
        "general": {
            "epochs": 1000,
            "n_dropout": 0.90,
            "adj_dropout": 0.05,
            "hidden_size": 128,
            "aug_adj": False,
            "improved": False,
            "n_pool": 4,
            "pool_rate": [0.7, 0.5, 0.5, 0.4],
        },
    },
    "ppnp": {
        "general": {},
        "flickr": {
            "lr": 0.005,
            "weight_decay": 0.001,
            "dropout": 0.3,
            "num_layers": 2,
            "hidden_size": 256,
            "epochs": 1000,
            "num_iterations": 4,
            "alpha": 0.5,
            # 0.5227
        },
        "cora": {
            "lr": 0.01,
            "weight_decay": 0.000887,
            "alpha": 0.1087,
            "num_iterations": 7,
            "num_layers": 2,
            "epochs": 1000,
            "hidden_size": 32,
            "dropout": 0.8185,
            #
        },
        "citeseer": {
            "num_iterations": 4,
            "hidden_size": 128,
            "lr": 0.01,
            "dropout": 0.3095,
            "alpha": 0.171,
            "weight_decay": 0.0004117,
            "num_layers": 2,
            "epochs": 1000,
            # 0.7199
        },
        "pubmed": {
            "num_iterations": 10,
            "hidden_size": 256,
            "dropout": 0.3766,
            "alpha": 0.1201,
            "weight_decay": 0.0002754,
            "epochs": 1000,
            "lr": 0.001,
        },
        "reddit": {
            "lr": 0.005,
            "hidden_size": 256,
            "dropout": 0.3751,
            "alpha": 0.1995,
            "num_layers": 3,
            "weight_decay": 5.918e-6,
            "num_iterations": 3,
            # 96.26
        },
    },
    "correct_smooth_mlp": {
        "general": {},
        "ogbn_arxiv": {
            "correct_norm": "row",
            "smooth_norm": "col",
            "correct_alpha": 0.9791632871592579,
            "smooth_alpha": 0.7564990804200602,
            "num_correct_prop": 50,
            "num_smooth_prop": 50,
            "autoscale": True,
            "norm": "batchnorm",
        },
        "ogbn_products": {
            "correct_norm": "sym",
            "smooth_norm": "row",
            "correct_alpha": 1.0,
            "smooth_alpha": 0.8,
            "num_correct_prop": 50,
            "num_smooth_prop": 50,
            "autoscale": False,
            "scale": 10.0,
            "norm": "batchnorm",
            "act_first": True,
        },
    },
    "sagn": {
        "general": {
            "data_gpu": True,
            "lr": 0.001,
            "hidden-size": 512,
            "attn-drop": 0.0,
            "dropout": 0.7,
        },
        "flickr": {
            "threshold": 0.5,
            "label-hop": 2,
            "weight-decay": 3e-6,
            "nstage": [50, 50, 50],
            "nhop": 2,
            "batch-size": 256,
        },
        "reddit": {
            "threshold": 0.9,
            "lr": 0.0001,
            "batch-size": 1000,
            "nhop": 2,
            "label-nhop": 4,
            "weight-decay": 0.0,
            "nstage": [500, 500, 500],
        },
    },
    "deepwalk": {
        "general": {},
    },
    "dngr": {
        "general": {
            "hidden_size": 128,
            "lr": 0.001,
            "epochs": 500,
            "hidden_size1": 1000,
            "hidden_size2": 128,
            "noise": 0.2,
            "alpha": 0.1,
            "step": 10,
        },
    },
    "grarep": {
        "general": {},
    },
    "hope": {
        "general": {},
    },
    "line": {
        "general": {},
        "blogcatalog": {"walk_num": 40},
    },
    "netmf": {
        "general": {},
        "ppi-ne": {"window_size": 10, "is_large": True},
        "blogcatalog": {"window_size": 10, "is_large": True},
        "wikipedia": {"window_size": 1},
    },
    "netsmf": {
        "general": {"window_size": 10, "num_round": 1000},
        "wikipedia": {"window_size": 1},
        "blogcatalog": {"num_round": 10000},
    },
    "node2vec": {
        "general": {},
    },
    "prone": {
        "general": {"step": 10},
        "ppi-ne": {"mu": 0.0},
        "wikipedia": {"mu": -4.0},
        "dblp-ne": {"mu": -1.2, "theta": 2.0},
    },
    "sdne": {
        "general": {},
    },
    "spectral": {
        "general": {},
    },
    "dgi": {
        "general": {"weight_decay": 0},
    },
    "gcc": {
        "general": {},
    },
    "grace": {
        "general": {
            "weight_decay": 0,
            "epochs": 1000,
            "patience": 20,
        },
        "cora": {
            "lr": 0.0005,
            "weight_decay": 0.00001,
            "tau": 0.4,
            "drop_feature_rates": [0.3, 0.4],
            "drop_edge_rates": [0.2, 0.4],
            "epochs": 200,
            "hidden_size": 128,
            "proj_hidden_size": 128,
        },
        "citeseer": {
            "hidden_size": 256,
            "proj_hidden_size": 256,
            "drop_feature_rates": [0.3, 0.2],
            "drop_edge_rates": [0.2, 0.0],
            "lr": 0.001,
            "_weight_decay": 0.00001,
            "tau": 0.9,
            "activation": "prelu",
        },
        "pubmed": {
            "hidden_size": 256,
            "proj_hidden_size": 256,
            "drop_edge_rates": [0.4, 0.1],
            "drop_feature_rates": [0.0, 0.2],
            "tau": 0.7,
            "lr": 0.001,
            "weight_decay": 0.00001,
        },
    },
    "unsup_graphsage": {
        "lr": 0.001,
        "weight_decay": 0,
        "epochs": 3000,
    },
    "revgcn": {
        "general": {},
        "cora": {
            "hidden_size": 128,
            "lr": 0.001,
            "dropout": 0.4706458854,
            "weight_decay": 0.0008907,
            "norm": "layernorm",
            "num_layers": 10,
            # 81.40
        },
    },
    "gin": {
        "general": {"lr": 0.001},
        "imdb-b": {"degree_feature": True},
        "imdb-m": {"degree_feature": True},
        "collab": {"degree_feature": True},
        "proteins": {
            "num_layers": 5,
            "dropout": 0.0,
        },
        "nci1": {
            "num_layers": 5,
            "dropout": 0.3,
            "hidden_size": 64,
        },
    },
    "infograph": {
        "general": {
            "lr": 0.0001,
            "weight_decay": 5e-4,
            "sup": False,
        },
        "mutag": {
            "num_layers": 1,
            "epochs": 20,
        },
        "imdb-b": {"degree_feature": True},
        "imdb-m": {"degree_feature": True},
        "collab": {"degree_feature": True},
        "nci1": {"num_layers": 3},
    },
    "sortpool": {
        "nci1": {
            "dropout": 0.3,
            "hidden_size": 64,
            "num_layers": 5,
        },
    },
    "patchy_san": {
        "general": {
            "lr": 0.001,
            "hidden_size": 32,
            "gamma": 0.5,
            "dropout": 0.5,
        },
        "imdb-b": {"degree_feature": True},
        "imdb-m": {"degree_feature": True},
        "collab": {"degree_feature": True},
    },
    "graph2vec": {
        "general": {},
        "nci1": {
            "lr": 0.001,
            "window_size": 8,
            "epochs": 10,
            "iteration": 4,
        },
        "reddit-b": {
            "lr": 0.01,
            "degree_feature": True,
            "hidden_size": 128,
        },
    },
    "gatne": {
        "general": {},
        "twitter": {"eval_type": "1"},
    },
    "hin2vec": {
        "general": {
            "lr": 0.025,
        },
    },
    "metapath2vec": {
        "general": {
            "walk_num": 40,
        },
    },
    "pte": {},
    "gtn": {
        "general": {
            "hidden_size": 128,
            "lr": 0.005,
            "weight_decay": 0.001,
        },
    },
    "han": {
        "general": {
            "hidden_size": 128,
            "lr": 0.005,
            "weight_decay": 0.001,
        }
    },
}



import os.path as osp

import torch

from cogdl.data import Graph, Dataset
from cogdl.utils import download_url


def read_gatne_data(folder):
    train_data = {}
    with open(osp.join(folder, "{}".format("train.txt")), "r") as f:
        for line in f:
            items = line.strip().split()
            if items[0] not in train_data:
                train_data[items[0]] = []
            train_data[items[0]].append([int(items[1]), int(items[2])])

    valid_data = {}
    with open(osp.join(folder, "{}".format("valid.txt")), "r") as f:
        for line in f:
            items = line.strip().split()
            if items[0] not in valid_data:
                valid_data[items[0]] = [[], []]
            valid_data[items[0]][1 - int(items[3])].append([int(items[1]), int(items[2])])

    test_data = {}
    with open(osp.join(folder, "{}".format("test.txt")), "r") as f:
        for line in f:
            items = line.strip().split()
            if items[0] not in test_data:
                test_data[items[0]] = [[], []]
            test_data[items[0]][1 - int(items[3])].append([int(items[1]), int(items[2])])

    data = Graph()
    data.train_data = train_data
    data.valid_data = valid_data
    data.test_data = test_data
    return data


class GatneDataset(Dataset):
    r"""The network datasets "Amazon", "Twitter" and "YouTube" from the
    `"Representation Learning for Attributed Multiplex Heterogeneous Network"
    <https://arxiv.org/abs/1905.01669>`_ paper.

    Args:
        root (string): Root directory where the dataset should be saved.
        name (string): The name of the dataset (:obj:`"Amazon"`,
            :obj:`"Twitter"`, :obj:`"YouTube"`).
    """

    url = "https://github.com/THUDM/GATNE/raw/master/data"

    def __init__(self, root, name):
        self.name = name
        super(GatneDataset, self).__init__(root)
        self.data = torch.load(self.processed_paths[0])

    @property
    def raw_file_names(self):
        names = ["train.txt", "valid.txt", "test.txt"]
        return names

    @property
    def processed_file_names(self):
        return ["data.pt"]

    def get(self, idx):
        assert idx == 0
        return self.data

    def download(self):
        for name in self.raw_file_names:
            download_url("{}/{}/{}".format(self.url, self.name.lower(), name), self.raw_dir)

    def process(self):
        data = read_gatne_data(self.raw_dir)
        torch.save(data, self.processed_paths[0])

    def __repr__(self):
        return "{}()".format(self.name)


class AmazonDataset(GatneDataset):
    def __init__(self, data_path="data"):
        dataset = "amazon"
        path = osp.join(data_path, dataset)
        super(AmazonDataset, self).__init__(path, dataset)


class TwitterDataset(GatneDataset):
    def __init__(self, data_path="data"):
        dataset = "twitter"
        path = osp.join(data_path, dataset)
        super(TwitterDataset, self).__init__(path, dataset)


class YouTubeDataset(GatneDataset):
    def __init__(self, data_path="data"):
        dataset = "youtube"
        path = osp.join(data_path, dataset)
        super(YouTubeDataset, self).__init__(path, dataset)



from .. import DataWrapper
from cogdl.wrappers.tools.wrapper_utils import node_degree_as_feature, split_dataset
from cogdl.data import DataLoader


class GraphClassificationDataWrapper(DataWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--degree-node-features", action="store_true",
                            help="Use one-hot degree vector as input node features")
        # parser.add_argument("--kfold", action="store_true", help="Use 10-fold cross-validation")
        parser.add_argument("--train-ratio", type=float, default=0.5)
        parser.add_argument("--test-ratio", type=float, default=0.3)
        parser.add_argument("--batch-size", type=int, default=16)
        # fmt: on

    def __init__(self, dataset, degree_node_features=False, batch_size=32, train_ratio=0.5, test_ratio=0.3):
        super(GraphClassificationDataWrapper, self).__init__(dataset)
        self.dataset = dataset
        self.degree_node_features = degree_node_features
        self.train_ratio = train_ratio
        self.test_ratio = test_ratio
        self.batch_size = batch_size
        self.split_idx = None

        self.setup_node_features()

    def train_wrapper(self):
        return DataLoader(self.dataset[self.split_idx[0]], batch_size=self.batch_size, shuffle=True, num_workers=4)

    def val_wrapper(self):
        if self.split_idx[1] is not None:
            return DataLoader(self.dataset[self.split_idx[1]], batch_size=self.batch_size, shuffle=False, num_workers=4)

    def test_wrapper(self):
        return DataLoader(self.dataset[self.split_idx[2]], batch_size=self.batch_size, shuffle=False, num_workers=4)

    def setup_node_features(self):
        if self.degree_node_features or self.dataset.data[0].x is None:
            self.dataset.data = node_degree_as_feature(self.dataset.data)
        train_idx, val_idx, test_idx = split_dataset(len(self.dataset), self.train_ratio, self.test_ratio)
        self.split_idx = [train_idx, val_idx, test_idx]



from .ops import (
    s_add_e,
    s_mul_e,
    s_sub_e,
    s_add_e_sum,
    s_mul_e_sum,
    s_sub_e_sum,
    s_add_e_mean,
    s_mul_e_mean,
    s_sub_e_mean,
    s_add_t,
    s_mul_t,
    s_dot_t,
    s_sub_t,
    s_div_t,
)



import torch
import torch.nn as nn

from cogdl.utils import edge_softmax


class DisenGCNLayer(nn.Module):
    """
    Implementation of `"Disentangled Graph Convolutional Networks" <http://proceedings.mlr.press/v97/ma19a.html>`_.
    """

    def __init__(self, in_feats, out_feats, K, iterations, tau=1.0, activation="leaky_relu"):
        super(DisenGCNLayer, self).__init__()
        self.K = K
        self.tau = tau
        self.iterations = iterations
        self.factor_dim = int(out_feats / K)

        self.weight = nn.Parameter(torch.Tensor(in_feats, out_feats))
        self.bias = nn.Parameter(torch.Tensor(out_feats))
        self.reset_parameters()

        if activation == "leaky_relu":
            self.activation = nn.LeakyReLU()
        elif activation == "sigmoid":
            self.activation = nn.Sigmoid()
        elif activation == "tanh":
            self.activation = nn.Tanh()
        elif activation == "prelu":
            self.activation = nn.PReLU()
        elif activation == "relu":
            self.activation = nn.ReLU()
        else:
            raise NotImplementedError

    def reset_parameters(self):
        nn.init.xavier_normal_(self.weight.data, gain=1.414)
        nn.init.zeros_(self.bias.data)

    def forward(self, graph, x):
        num_nodes = x.shape[0]
        device = x.device

        h = self.activation(torch.matmul(x, self.weight) + self.bias)

        h = h.split(self.factor_dim, dim=-1)
        h = torch.cat([dt.unsqueeze(0) for dt in h], dim=0)
        norm = h.pow(2).sum(dim=-1).sqrt().unsqueeze(-1)

        # multi-channel softmax: faster
        h_normed = h / norm  # (K, N, d)
        h_src = h_dst = h_normed.permute(1, 0, 2)  # (N, K, d)
        add_shape = h.shape  # (K, N, d)

        edge_index = graph.edge_index
        for _ in range(self.iterations):
            src_edge_attr = h_dst[edge_index[0]] * h_src[edge_index[1]]
            src_edge_attr = src_edge_attr.sum(dim=-1) / self.tau  # shape: (N, K)
            edge_attr_softmax = edge_softmax(graph, src_edge_attr).T  # shape: (E, K)
            edge_attr_softmax = edge_attr_softmax.unsqueeze(-1)  # shape: (K, E, 1)

            dst_edge_attr = h_src.index_select(0, edge_index[1]).permute(1, 0, 2)  # shape: (E, K, d) -> (K, E, d)
            dst_edge_attr = dst_edge_attr * edge_attr_softmax
            edge_index_ = edge_index[0].unsqueeze(-1).unsqueeze(0).repeat(self.K, 1, h.shape[-1])
            node_attr = torch.zeros(add_shape).to(device).scatter_add_(1, edge_index_, dst_edge_attr)  # (K, N, d)
            node_attr = node_attr + h_normed
            node_attr_norm = node_attr.pow(2).sum(-1).sqrt().unsqueeze(-1)  # shape: (K, N, 1)
            node_attr = (node_attr / node_attr_norm).permute(1, 0, 2)  # shape: (N, K, d)
            h_dst = node_attr

        h_dst = h_dst.reshape(num_nodes, -1)

        # Calculate the softmax of each channel separately
        # h_src = h_dst = h / norm  # (K, N, d)
        #
        # for _ in range(self.iterations):
        #     for i in range(self.K):
        #         h_attr = h_dst[i]
        #         edge_attr = h_attr[edge_index[0]] * h_src[i][edge_index[1]]
        #
        #         edge_attr = edge_attr.sum(-1)/self.tau
        #         edge_attr = edge_softmax(edge_index, edge_attr, shape=(num_nodes, num_nodes))
        #
        #         node_attr = spmm(edge_index, edge_attr, h_src[i])
        #
        #         node_attr = node_attr + h_src[i]
        #         h_src[i] = node_attr / node_attr.pow(2).sum(-1).sqrt().unsqueeze(-1)
        #
        # h_dst = h_dst.permute(1, 0, 2).reshape(num_nodes, -1)

        return h_dst



import random
import argparse
import networkx as nx
import numpy as np
from gensim.models import Word2Vec
from tqdm import tqdm

from .. import BaseModel


class DeepWalk(BaseModel):
    r"""The DeepWalk model from the `"DeepWalk: Online Learning of Social Representations"
    <https://arxiv.org/abs/1403.6652>`_ paper

    Args:
        hidden_size (int) : The dimension of node representation.
        walk_length (int) : The walk length.
        walk_num (int) : The number of walks to sample for each node.
        window_size (int) : The actual context size which is considered in language model.
        worker (int) : The number of workers for word2vec.
        iteration (int) : The number of training iteration in word2vec.
    """

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--walk-length", type=int, default=80,
                            help="Length of walk per source. Default is 80.")
        parser.add_argument("--walk-num", type=int, default=40,
                            help="Number of walks per source. Default is 40.")
        parser.add_argument("--window-size", type=int, default=5,
                            help="Window size of skip-gram model. Default is 5.")
        parser.add_argument("--worker", type=int, default=10,
                            help="Number of parallel workers. Default is 10.")
        parser.add_argument("--iteration", type=int, default=10,
                            help="Number of iterations. Default is 10.")
        parser.add_argument("--hidden-size", type=int, default=128)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args) -> "DeepWalk":
        return cls(
            args.hidden_size,
            args.walk_length,
            args.walk_num,
            args.window_size,
            args.worker,
            args.iteration,
        )

    def __init__(self, dimension, walk_length, walk_num, window_size, worker, iteration):
        super(DeepWalk, self).__init__()
        self.dimension = dimension
        self.walk_length = walk_length
        self.walk_num = walk_num
        self.window_size = window_size
        self.worker = worker
        self.iteration = iteration

    def forward(self, graph, embedding_model_creator=Word2Vec, return_dict=False):
        nx_g = graph.to_networkx()
        self.G = nx_g
        walks = self._simulate_walks(self.walk_length, self.walk_num)
        walks = [[str(node) for node in walk] for walk in walks]
        print("training word2vec...")
        model = embedding_model_creator(
            walks,
            size=self.dimension,
            window=self.window_size,
            min_count=0,
            sg=1,
            workers=self.worker,
            iter=self.iteration,
        )
        id2node = dict([(vid, node) for vid, node in enumerate(nx_g.nodes())])
        embeddings = np.asarray([model.wv[str(id2node[i])] for i in range(len(id2node))])

        if return_dict:
            features_matrix = dict()
            for vid, node in enumerate(nx_g.nodes()):
                features_matrix[node] = embeddings[vid]
        else:
            features_matrix = np.zeros((graph.num_nodes, embeddings.shape[1]))
            nx_nodes = nx_g.nodes()
            features_matrix[nx_nodes] = embeddings[np.arange(graph.num_nodes)]
        return features_matrix

    def _walk(self, start_node, walk_length):
        # Simulate a random walk starting from start node.
        walk = [start_node]
        while len(walk) < walk_length:
            cur = walk[-1]
            cur_nbrs = list(self.G.neighbors(cur))
            if len(cur_nbrs) == 0:
                break
            k = int(np.floor(np.random.rand() * len(cur_nbrs)))
            walk.append(cur_nbrs[k])
        return walk

    def _simulate_walks(self, walk_length, num_walks):
        # Repeatedly simulate random walks from each node.
        G = self.G
        walks = []
        nodes = list(G.nodes())
        print("node number:", len(nodes))
        print("generating random walks...")
        for walk_iter in tqdm(range(num_walks)):
            random.shuffle(nodes)
            for node in nodes:
                walks.append(self._walk(node, walk_length))
        return walks



import torch
import torch.nn as nn
from cogdl.layers import QGCNLayer

from .. import BaseModel


class QGCN(BaseModel):
    r"""The GCN model from the `"Semi-Supervised Classification with Graph Convolutional Networks"
    <https://arxiv.org/abs/1609.02907>`_ paper

    Args:
        in_features (int) : Number of input features.
        out_features (int) : Number of classes.
        hidden_size (int) : The dimension of node representation.
        dropout (float) : Dropout rate for model training.
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--residual", action="store_true")
        parser.add_argument("--norm", type=str, default=None)
        parser.add_argument("--activation", type=str, default="relu")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.num_layers,
            args.dropout,
            args.activation,
            args.residual,
            args.norm,
            args.rp_ratio,
        )

    def __init__(
        self,
        in_feats,
        hidden_size,
        out_feats,
        num_layers,
        dropout,
        activation="relu",
        residual=False,
        norm=None,
        rp_ratio=1,
    ):
        super(QGCN, self).__init__()
        shapes = [in_feats] + [hidden_size] * (num_layers - 1) + [out_feats]
        self.layers = nn.ModuleList(
            [
                QGCNLayer(
                    shapes[i],
                    shapes[i + 1],
                    dropout=dropout if i != num_layers - 1 else 0,
                    residual=residual if i != num_layers - 1 else None,
                    norm=norm if i != num_layers - 1 else None,
                    activation=activation if i != num_layers - 1 else None,
                    rp_ratio=rp_ratio,
                )
                for i in range(num_layers)
            ]
        )
        self.quant = torch.quantization.QuantStub()
        self.dequant = torch.quantization.DeQuantStub()
        self.num_layers = num_layers

    def embed(self, graph):
        graph.sym_norm()
        h = graph.x
        for i in range(self.num_layers - 1):
            h = self.layers[i](graph, h)
        return h

    def forward(self, graph):
        graph.sym_norm()
        h = graph.x
        h = self.quant(h)
        for i in range(self.num_layers):
            h = self.layers[i](graph, h)
        h = self.dequant(h)
        return h

    def predict(self, data):
        return self.forward(data)



import torch

import numpy as np
from cogdl.utils import RandomWalker
from cogdl.wrappers.tools.wrapper_utils import evaluate_node_embeddings_using_logreg
from .. import ModelWrapper


class UnsupGraphSAGEModelWrapper(ModelWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--walk-length", type=int, default=10)
        parser.add_argument("--negative-samples", type=int, default=30)
        # fmt: on

    def __init__(self, model, optimizer_cfg, walk_length, negative_samples):
        super(UnsupGraphSAGEModelWrapper, self).__init__()
        self.model = model
        self.optimizer_cfg = optimizer_cfg
        self.walk_length = walk_length
        self.num_negative_samples = negative_samples
        self.random_walker = RandomWalker()

    def train_step(self, batch):
        data = batch
        x = self.model(data)
        device = x.device

        self.random_walker.build_up(data.edge_index, data.x.shape[0])
        walk_res = self.random_walker.walk(
            start=torch.arange(0, x.shape[0]).to(device), walk_length=self.walk_length + 1
        )
        self.walk_res = torch.as_tensor(walk_res)[:, 1:]

        self.num_nodes = max(data.edge_index[0].max(), data.edge_index[1].max()).item() + 1

        self.negative_samples = torch.from_numpy(
            np.random.choice(self.num_nodes, (self.num_nodes, self.num_negative_samples))
        ).to(device)

        pos_loss = -torch.log(
            torch.sigmoid(torch.sum(x.unsqueeze(1).repeat(1, self.walk_length, 1) * x[self.walk_res], dim=-1))
        ).mean()
        neg_loss = -torch.log(
            torch.sigmoid(
                -torch.sum(x.unsqueeze(1).repeat(1, self.num_negative_samples, 1) * x[self.negative_samples], dim=-1)
            )
        ).mean()
        return (pos_loss + neg_loss) / 2

    def test_step(self, graph):
        with torch.no_grad():
            pred = self.model(graph)
        y = graph.y
        result = evaluate_node_embeddings_using_logreg(pred, y, graph.train_mask, graph.test_mask)
        self.note("test_acc", result)

    def setup_optimizer(self):
        cfg = self.optimizer_cfg
        return torch.optim.Adam(self.parameters(), lr=cfg["lr"], weight_decay=cfg["weight_decay"])



import os

import torch
from sklearn.preprocessing import StandardScaler

from cogdl.data import Dataset, Graph, MultiGraphDataset
from cogdl.utils import Accuracy, MultiLabelMicroF1, MultiClassMicroF1, CrossEntropyLoss, BCEWithLogitsLoss


def _get_evaluator(metric):
    if metric == "accuracy":
        return Accuracy()
    elif metric == "multilabel_f1":
        return MultiLabelMicroF1()
    elif metric == "multiclass_f1":
        return MultiClassMicroF1()
    else:
        raise NotImplementedError


def _get_loss_fn(metric):
    if metric in ("accuracy", "multiclass_f1"):
        return CrossEntropyLoss()
    elif metric == "multilabel_f1":
        return BCEWithLogitsLoss()
    else:
        raise NotImplementedError


def scale_feats(data):
    scaler = StandardScaler()
    feats = data.x.numpy()
    scaler.fit(feats)
    feats = torch.from_numpy(scaler.transform(feats)).float()
    data.x = feats
    return data


def generate_random_graph(num_nodes=100, num_edges=1000, num_feats=64):
    # load or generate your dataset
    edge_index = torch.randint(0, num_nodes, (2, num_edges))
    x = torch.randn(num_nodes, num_feats)
    y = torch.randint(0, 2, (num_nodes,))

    # set train/val/test mask in node_classification task
    train_mask = torch.zeros(num_nodes).bool()
    train_mask[0 : int(0.3 * num_nodes)] = True
    val_mask = torch.zeros(num_nodes).bool()
    val_mask[int(0.3 * num_nodes) : int(0.7 * num_nodes)] = True
    test_mask = torch.zeros(num_nodes).bool()
    test_mask[int(0.7 * num_nodes) :] = True
    data = Graph(x=x, edge_index=edge_index, y=y, train_mask=train_mask, val_mask=val_mask, test_mask=test_mask)

    return data


class NodeDataset(Dataset):
    """
    data_path : path to load dataset. The dataset must be processed to specific format
    metric: Accuracy, multi-label f1 or multi-class f1. Default: `accuracy`
    """

    def __init__(self, path="data.pt", data=None, scale_feat=True, metric="auto"):
        self.path = path
        self.data = data
        super(NodeDataset, self).__init__(root=path)
        if self.data is None:
            try:
                self.data = torch.load(path)
            except Exception as e:
                print(e)
                exit(1)
        if scale_feat:
            self.data = scale_feats(self.data)
        self.metric = metric
        if hasattr(self.data, "y") and self.data.y is not None:
            if metric == "auto":
                if len(self.data.y.shape) > 1:
                    self.metric = "multilabel_f1"
                else:
                    self.metric = "accuracy"

    def download(self):
        pass

    def process(self):
        if self.data is None:
            raise NotImplementedError
        return self.data

    def get(self, idx):
        assert idx == 0
        return self.data

    def get_evaluator(self):
        return _get_evaluator(self.metric)

    def get_loss_fn(self):
        return _get_loss_fn(self.metric)

    def _download(self):
        pass

    def _process(self):
        if not os.path.exists(self.path):
            data = self.process()
            torch.save(data, self.path)

    def __repr__(self):
        return "{}".format(self.path)


class GraphDataset(MultiGraphDataset):
    def __init__(self, path="cus_graph_data.pt", metric="accuracy"):
        self.path = path
        super(GraphDataset, self).__init__(root=path)
        # try:
        data = torch.load(path)
        if hasattr(data[0], "y") and data[0].y is None:
            self.y = torch.cat([idata.y for idata in data])
        self.data = data

        self.metric = metric
        if hasattr(self.data, "y") and self.data.y is not None:
            if metric == "auto":
                if len(self.data.y.shape) > 1:
                    self.metric = "multilabel_f1"
                else:
                    self.metric = "accuracy"

    def _download(self):
        pass

    def process(self):
        raise NotImplementedError

    def _process(self):
        if not os.path.exists(self.path):
            data = self.process()
            torch.save(data, self.path)

    def get_evaluator(self):
        return _get_evaluator(self.metric)

    def get_loss_fn(self):
        return _get_loss_fn(self.metric)

    def __repr__(self):
        return "{}".format(self.path)



import torch
import torch.nn.functional as F

from .. import BaseModel
from cogdl.utils import spmm
from .mlp import MLP


class PPNP(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--propagation-type", type=str, default="appnp")
        parser.add_argument("--alpha", type=float, default=0.1)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--num-iterations", type=int, default=10)  # only for appnp
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.num_layers,
            args.dropout,
            args.propagation_type,
            args.alpha,
            args.num_iterations,
        )

    def __init__(self, nfeat, nhid, nclass, num_layers, dropout, propagation, alpha, niter, cache=True):
        super(PPNP, self).__init__()
        # GCN as a prediction and then apply the personalized page rank on the results
        self.nn = MLP(nfeat, nclass, nhid, num_layers, dropout)
        if propagation not in ("appnp", "ppnp"):
            print("Invalid propagation type, using default appnp")
            propagation = "appnp"

        self.propagation = propagation
        self.alpha = alpha
        self.niter = niter
        self.dropout = dropout
        self.vals = None  # speedup for ppnp
        self.use_cache = cache
        self.cache = dict()

    def forward(self, graph):
        def get_ready_format(input, edge_index, edge_attr=None):
            if isinstance(edge_index, tuple):
                edge_index = torch.stack(edge_index)
            if edge_attr is None:
                edge_attr = torch.ones(edge_index.shape[1]).float().to(input.device)
            adj = torch.sparse_coo_tensor(
                edge_index,
                edge_attr,
                (input.shape[0], input.shape[0]),
            ).to(input.device)
            return adj

        x = graph.x
        graph.sym_norm()
        # get prediction
        x = F.dropout(x, p=self.dropout, training=self.training)
        local_preds = self.nn.forward(x)

        # apply personalized pagerank
        if self.propagation == "ppnp":
            if self.vals is None:
                self.vals = self.alpha * torch.inverse(
                    torch.eye(x.shape[0]).to(x.device)
                    - (1 - self.alpha) * get_ready_format(x, graph.edge_index, graph.edge_attr)
                )
            final_preds = F.dropout(self.vals) @ local_preds
        else:  # appnp
            preds = local_preds
            with graph.local_graph():
                graph.edge_weight = F.dropout(graph.edge_weight, p=self.dropout, training=self.training)
                graph.set_symmetric()
                for _ in range(self.niter):
                    new_features = spmm(graph, preds)
                    preds = (1 - self.alpha) * new_features + self.alpha * local_preds
                final_preds = preds
        return final_preds

    def predict(self, graph):
        return self.forward(graph)



from typing import Dict

import random
import numpy as np
import scipy.sparse as sp
from collections import defaultdict

import torch
import torch.nn as nn
import torch.nn.functional as F

from sklearn.utils import shuffle as skshuffle
from sklearn.linear_model import LogisticRegression
from sklearn.multiclass import OneVsRestClassifier
from sklearn.model_selection import GridSearchCV, KFold
from sklearn.svm import SVC
from sklearn.cluster import KMeans, SpectralClustering
from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.metrics import f1_score
from scipy.optimize import linear_sum_assignment

from cogdl.utils import accuracy, multilabel_f1


def pre_evaluation_index(y_pred, y_true, sigmoid=False):
    """
    Pre-calculating diffusion matrix for mini-batch evaluation
    Return:
        torch.Tensor((tp, all)) for multi-class classification
        torch.Tensor((tp, fp, fn)) for multi-label classification
    """
    if len(y_true.shape) == 1:
        pred = (y_pred.argmax(1) == y_true).int()
        tp = pred.sum()
        fnp = pred.shape[0] - tp
        return torch.tensor((tp, fnp)).float()
    else:
        if sigmoid:
            border = 0.5
        else:
            border = 0
        y_pred[y_pred >= border] = 1
        y_pred[y_pred < border] = 0
        tp = (y_pred * y_true).sum().to(torch.float32)
        # tn = ((1 - y_true) * (1 - y_pred)).sum().to(torch.float32)
        fp = ((1 - y_true) * y_pred).sum().to(torch.float32)
        fn = (y_true * (1 - y_pred)).sum().to(torch.float32)
        return torch.tensor((tp, fp, fn))


def merge_batch_indexes(values: list, method="mean"):
    # if key.endswith("loss"):
    #     result = sum(values)
    #     if torch.is_tensor(result):
    #         result = result.item()
    #     result = result / len(values)
    # elif key.endswith("eval_index"):
    #     if len(values) > 1:
    #         val = torch.stack(values)
    #         val = val.sum(0)
    #     else:
    #         val = values[0]
    #     fp = val[0]
    #     all_ = val.sum()
    #
    #     prefix = key[: key.find("eval_index")]
    #     if val.shape[0] == 2:
    #         _key = prefix + "acc"
    #     else:
    #         _key = prefix + "f1"
    #     result = (fp / all_).item()
    #     out_key = _key

    if isinstance(values[0], dict) or isinstance(values[0], tuple):
        return values
    elif method == "mean":
        return sum(values) / len(values)
    elif method == "sum":
        return sum(values)
    else:
        return sum(values)


def node_degree_as_feature(data):
    r"""
    Set each node feature as one-hot encoding of degree
    :param data: a list of class Data
    :return: a list of class Data
    """
    max_degree = 0
    degrees = []
    device = data[0].edge_index[0].device

    for graph in data:
        deg = graph.degrees().long()
        degrees.append(deg)
        max_degree = max(deg.max().item(), max_degree)

    max_degree = int(max_degree) + 1
    for i in range(len(data)):
        one_hot = F.one_hot(degrees[i], max_degree).float()
        data[i].x = one_hot.to(device)
    return data


def split_dataset(ndata, train_ratio, test_ratio):

    train_size = int(ndata * train_ratio)
    test_size = int(ndata * test_ratio)
    index = np.arange(ndata)
    random.shuffle(index)

    train_index = index[:train_size]
    test_index = index[-test_size:]
    if train_ratio + test_ratio == 1:
        val_index = None
    else:
        val_index = index[train_size:-test_size]
    return train_index, val_index, test_index


def evaluate_node_embeddings_using_logreg(data, labels, train_idx, test_idx, run=20):
    result = LogRegTrainer().train(data, labels, train_idx, test_idx, run=run)
    return result


class LogReg(nn.Module):
    def __init__(self, ft_in, nb_classes):
        super(LogReg, self).__init__()
        self.fc = nn.Linear(ft_in, nb_classes)

        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    def forward(self, seq):
        ret = self.fc(seq)
        return ret


class LogRegTrainer(object):
    def train(self, data, labels, idx_train, idx_test, loss_fn=None, evaluator=None, run=20):
        device = data.device
        nhid = data.shape[-1]
        labels = labels.to(device)

        train_embs = data[idx_train]
        test_embs = data[idx_test]

        train_lbls = labels[idx_train]
        test_lbls = labels[idx_test]
        tot = 0

        num_classes = int(labels.max()) + 1

        if loss_fn is None:
            loss_fn = nn.CrossEntropyLoss() if len(labels.shape) == 1 else nn.BCEWithLogitsLoss()

        if evaluator is None:
            evaluator = accuracy if len(labels.shape) == 1 else multilabel_f1

        for _ in range(run):
            log = LogReg(nhid, num_classes).to(device)
            optimizer = torch.optim.Adam(log.parameters(), lr=0.01, weight_decay=0.0)
            log.to(device)

            for _ in range(100):
                log.train()
                optimizer.zero_grad()

                logits = log(train_embs)
                loss = loss_fn(logits, train_lbls)

                loss.backward()
                optimizer.step()

            log.eval()
            with torch.no_grad():
                logits = log(test_embs)
            metric = evaluator(logits, test_lbls)

            tot += metric
        return tot / run


def evaluate_node_embeddings_using_liblinear(features_matrix, label_matrix, num_shuffle, training_percents):
    if len(label_matrix.shape) > 1:
        labeled_nodes = np.nonzero(np.sum(label_matrix, axis=1) > 0)[0]
        features_matrix = features_matrix[labeled_nodes]
        label_matrix = label_matrix[labeled_nodes]

    # shuffle, to create train/test groups
    shuffles = []
    for _ in range(num_shuffle):
        shuffles.append(skshuffle(features_matrix, label_matrix))

    # score each train/test group
    all_results = defaultdict(list)

    for train_percent in training_percents:
        for shuf in shuffles:
            X, y = shuf

            training_size = int(train_percent * len(features_matrix))
            X_train = X[:training_size, :]
            y_train = y[:training_size, :]

            X_test = X[training_size:, :]
            y_test = y[training_size:, :]

            clf = TopKRanker(LogisticRegression(solver="liblinear"))
            clf.fit(X_train, y_train)

            # find out how many labels should be predicted
            top_k_list = y_test.sum(axis=1).astype(np.int).tolist()
            preds = clf.predict(X_test, top_k_list)
            result = f1_score(y_test, preds, average="micro")
            all_results[train_percent].append(result)

    return dict(
        (f"Micro-F1 {train_percent}", np.mean(all_results[train_percent]))
        for train_percent in sorted(all_results.keys())
    )


class TopKRanker(OneVsRestClassifier):
    def predict(self, X, top_k_list):
        assert X.shape[0] == len(top_k_list)
        probs = np.asarray(super(TopKRanker, self).predict_proba(X))
        all_labels = sp.lil_matrix(probs.shape)

        for i, k in enumerate(top_k_list):
            probs_ = probs[i, :]
            labels = self.classes_[probs_.argsort()[-k:]].tolist()
            for label in labels:
                all_labels[i, label] = 1
        return all_labels


def evaluate_graph_embeddings_using_svm(embeddings, labels):
    result = []
    kf = KFold(n_splits=10)
    kf.get_n_splits(X=embeddings, y=labels)
    for train_index, test_index in kf.split(embeddings):
        x_train = embeddings[train_index]
        x_test = embeddings[test_index]
        y_train = labels[train_index]
        y_test = labels[test_index]
        params = {"C": [1e-2, 1e-1, 1]}
        svc = SVC()
        clf = GridSearchCV(svc, params)
        clf.fit(x_train, y_train)

        preds = clf.predict(x_test)
        f1 = f1_score(y_test, preds, average="micro")
        result.append(f1)
    test_f1 = np.mean(result)
    test_std = np.std(result)

    return dict(acc=test_f1, std=test_std)


def evaluate_clustering(features_matrix, labels, cluster_method, num_clusters, num_nodes, full=True):
    print("Clustering...")
    if cluster_method == "kmeans":
        kmeans = KMeans(n_clusters=num_clusters, random_state=0).fit(features_matrix)
        clusters = kmeans.labels_
    else:
        clustering = SpectralClustering(n_clusters=num_clusters, assign_labels="discretize", random_state=0).fit(
            features_matrix
        )
        clusters = clustering.labels_

    print("Evaluating...")
    truth = labels.cpu().numpy()
    if full:
        mat = np.zeros([num_clusters, num_clusters])
        for i in range(num_nodes):
            mat[clusters[i]][truth[i]] -= 1
        _, row_idx = linear_sum_assignment(mat)
        acc = -mat[_, row_idx].sum() / num_nodes
        for i in range(num_nodes):
            clusters[i] = row_idx[clusters[i]]
        macro_f1 = f1_score(truth, clusters, average="macro")
        return dict(acc=acc, nmi=normalized_mutual_info_score(clusters, truth), macro_f1=macro_f1)
    else:
        return dict(nmi=normalized_mutual_info_score(clusters, truth))



import os
import re
import torch
import numpy as np
from .dual_position_bert_model import DualPositionBertForPreTrainingPreLN
from .utils import colored, OAG_TOKEN_TYPE_NAMES, stringLenCJK, stringRjustCJK
from transformers import BertTokenizer
import sentencepiece as spm


class OAGMetaInfoBertModel(DualPositionBertForPreTrainingPreLN):
    def __init__(self, bert_config, tokenizer):
        super(OAGMetaInfoBertModel, self).__init__(bert_config)
        self.tokenizer = tokenizer
        self.spm = not isinstance(self.tokenizer, BertTokenizer)
        if self.spm:
            (
                self.tokenizer.cls_token_id,
                self.tokenizer.mask_token_id,
                self.tokenizer.sep_token_id,
            ) = self.tokenizer.PieceToId(["[CLS]", "[MASK]", "[SEP]"])

    def __recursively_build_spm_token_ids(self, text, splitters=[]):
        """
        SentencePiece tokenizer cannot directly decode control symbols such as [MASK] or [SEP]. This function will handle this problem.
        """
        if len(splitters) == 0:
            return self.tokenizer.encode(text)
        splitter = splitters[0]
        splitters = splitters[1:]
        start = 0
        parts = []
        while start >= 0 and start < len(text):
            end = text.find(splitter, start)
            if end >= 0:
                parts += self.__recursively_build_spm_token_ids(text[start:end].strip(), splitters)
                start = end + len(splitter)
                parts.append(self.tokenizer.PieceToId(splitter))
            else:
                end = len(text)
                parts += self.__recursively_build_spm_token_ids(text[start:end].strip(), splitters)
                break
        return parts

    def _convert_text_to_token_ids(self, text):
        if self.spm:
            return self.__recursively_build_spm_token_ids(
                text, splitters=["[PAD]", "[EOS]", "[UNK]", "[CLS]", "[SEP]", "[MASK]", "[BOS]"]
            )
        else:
            return self.tokenizer(text, add_special_tokens=False)["input_ids"] if len(text) > 0 else []

    def _convert_ids_to_tokens(self, token_ids):
        if self.spm:
            _ids = []
            for _id in token_ids:
                if not isinstance(_id, int):
                    _id = _id.item()
                _ids.append(_id)
            return self.tokenizer.id_to_piece(_ids)
        else:
            return self.tokenizer.convert_ids_to_tokens(token_ids)

    def _convert_token_ids_to_text(self, token_ids):
        if self.spm:
            _ids = []
            for _id in token_ids:
                if not isinstance(_id, int):
                    _id = _id.item()
                _ids.append(_id)
            return self.tokenizer.decode(_ids)
        else:
            return self.tokenizer.convert_tokens_to_string(self.tokenizer.convert_ids_to_tokens(token_ids))

    def print_oag_instance(
        self,
        input_ids,
        token_type_ids,
        input_masks,
        masked_lm_labels,
        position_ids,
        position_ids_second,
        predictions=None,
    ):
        COLORS = ["white", "green", "blue", "red", "yellow", "magenta"]
        try:
            termwidth, _ = os.get_terminal_size()
        except Exception:
            termwidth = 200
        K = predictions.shape[1] if predictions is not None else 0
        input_ids = [token_id for i, token_id in enumerate(input_ids) if input_masks[i].sum() > 0]
        position_ids = [position_id for i, position_id in enumerate(position_ids) if input_masks[i].sum() > 0]
        position_ids_second = [
            position_id for i, position_id in enumerate(position_ids_second) if input_masks[i].sum() > 0
        ]
        token_type_ids = [token_type_id for i, token_type_id in enumerate(token_type_ids) if input_masks[i].sum() > 0]
        masks = [0 for i in input_ids]
        prediction_topks = [[0 for i in input_ids] for _ in range(K)]
        mask_indices = []
        for lm_pos, lm_id in enumerate(masked_lm_labels):
            if lm_id < 0:
                continue
            masks[lm_pos] = lm_id
            mask_indices.append(lm_pos)
        for k in range(K):
            for lm_pos, token_id in zip(mask_indices, predictions[:, k]):
                prediction_topks[k][lm_pos] = token_id
        input_tokens = self._convert_ids_to_tokens(input_ids)
        masks_tokens = self._convert_ids_to_tokens(masks)
        prediction_tokens = [self._convert_ids_to_tokens(prediction_topks[k]) for k in range(K)]
        input_tokens_str = [""]
        position_ids_str = [""]
        position_ids_second_str = [""]
        token_type_ids_str = [""]
        masks_str = [""]
        prediction_topk_strs = [[""] for _ in range(K)]
        current_length = 0
        for pos, (input_token, position_id, position_id_second, token_type_id, mask) in enumerate(
            zip(input_tokens, position_ids, position_ids_second, token_type_ids, masks_tokens)
        ):
            token_type = OAG_TOKEN_TYPE_NAMES[token_type_id]
            length = max(
                stringLenCJK(input_token) + 1,
                7,
                stringLenCJK(token_type) + 1,
                stringLenCJK(mask) + 1,
                *[stringLenCJK(prediction_tokens[k][pos]) + 1 for k in range(K)],
            )
            if current_length + length > termwidth:
                current_length = 0
                input_tokens_str.append("")
                position_ids_str.append("")
                position_ids_second_str.append("")
                token_type_ids_str.append("")
                masks_str.append("")
                for k in range(K):
                    prediction_topk_strs[k].append("")
            current_length += length
            input_tokens_str[-1] += colored(stringRjustCJK(input_token, length), COLORS[token_type_id])
            position_ids_str[-1] += stringRjustCJK(str(position_id), length)
            position_ids_second_str[-1] += stringRjustCJK(str(position_id_second), length)
            token_type_ids_str[-1] += stringRjustCJK(token_type, length)
            masks_str[-1] += colored(
                stringRjustCJK(mask, length) if mask != "[PAD]" else stringRjustCJK("", length), COLORS[token_type_id]
            )
            for k in range(K):
                v = prediction_tokens[k][pos] if prediction_tokens[k][pos] != "[PAD]" else ""
                prediction_topk_strs[k][-1] += colored(
                    stringRjustCJK(v, length), "magenta" if v != mask and mask != "[CLS]" else "cyan"
                )

        ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        size = len(masks_str)
        for i in range(size):
            print()
            print(input_tokens_str[i])
            print(position_ids_str[i])
            print(position_ids_second_str[i])
            print(token_type_ids_str[i])
            if ansi_escape.sub("", masks_str[i]).strip() != "":
                print(masks_str[i])
            for k in range(K):
                if ansi_escape.sub("", prediction_topk_strs[k][i]).strip() != "":
                    print(prediction_topk_strs[k][i])
            print("-" * termwidth)

    def build_inputs(
        self,
        title="",
        abstract="",
        venue="",
        authors=[],
        concepts=[],
        affiliations=[],
        decode_span_type="FOS",
        decode_span_length=0,
        max_seq_length=512,
        mask_propmt_text="",
    ):
        """build inputs from text information for model to use

        Args:
            title (str, optional): [paper title]. Defaults to ''.
            abstract (str, optional): [paper abstract]. Defaults to ''.
            venue (str, optional): [paper venue]. Defaults to ''.
            authors (list, optional): [paper author]. Defaults to [].
            concepts (list, optional): [paper concepts]. Defaults to [].
            affiliations (list, optional): [paper affiliations]. Defaults to [].
            decode_span_type (str, optional): [the span type to decode, choose from 'FOS','VENUE','AFF','AUTHOR']. Defaults to 'FOS'.
            decode_span_length (int, optional): [the length of span to decode]. Defaults to 0.
            max_seq_length (int, optional): [maximum sequence length for the input, the context information will be truncated if the total length exceeds this number]. Defaults to 512.
            mask_propmt_text (str, optional): [the prompt text to add after title and abstract]. Defaults to ''.

        Raises:
            Exception: [provided inputs are invalid]

        Returns:
            [tuple of list]: [input_ids, input_masks, token_type_ids, masked_lm_labels, position_ids, position_ids_second, masked_positions, num_spans]
        """
        input_ids = []
        input_masks = []
        token_type_ids = []
        masked_lm_labels = []
        position_ids = []
        position_ids_second = []
        masked_positions = []
        num_spans = 0

        def add_span(token_type_id, token_ids, is_mask=False):
            nonlocal num_spans
            if len(token_ids) == 0:
                return
            length = len(token_ids)
            input_ids.extend(token_ids if not is_mask else [self.tokenizer.mask_token_id] * length)
            input_masks.extend([1] * length)
            token_type_ids.extend([token_type_id] * length)
            masked_lm_labels.extend([-1] * length if not is_mask else [self.tokenizer.cls_token_id] * length)
            position_ids.extend([num_spans] * length)
            position_ids_second.extend(list(range(length)))
            if is_mask:
                masked_positions.extend([len(input_ids) - length + i for i in range(decode_span_length)])
            num_spans += 1

        span_token_type_id = OAG_TOKEN_TYPE_NAMES.index(decode_span_type)
        if span_token_type_id < 0:
            raise Exception("Unexpected span type: %s" % decode_span_type)

        prompt_token_ids = self._convert_text_to_token_ids(mask_propmt_text)

        add_span(
            0,
            (self._convert_text_to_token_ids(title) + self._convert_text_to_token_ids(abstract) + prompt_token_ids)[
                : max_seq_length - decode_span_length
            ],
        )
        add_span(2, self._convert_text_to_token_ids(venue)[: max_seq_length - len(input_ids) - decode_span_length])
        for author in authors:
            add_span(1, self._convert_text_to_token_ids(author)[: max_seq_length - len(input_ids) - decode_span_length])
        for concept in concepts:
            add_span(
                4, self._convert_text_to_token_ids(concept)[: max_seq_length - len(input_ids) - decode_span_length]
            )
        for affiliation in affiliations:
            add_span(
                3, self._convert_text_to_token_ids(affiliation)[: max_seq_length - len(input_ids) - decode_span_length]
            )

        add_span(span_token_type_id, [0 for _ in range(decode_span_length)], is_mask=True)
        return (
            input_ids,
            input_masks,
            token_type_ids,
            masked_lm_labels,
            position_ids,
            position_ids_second,
            masked_positions,
            num_spans,
        )

    def encode_paper(
        self,
        title="",
        abstract="",
        venue="",
        authors=[],
        concepts=[],
        affiliations=[],
        decode_span_type="FOS",
        decode_span_length=0,
        max_seq_length=512,
        mask_propmt_text="",
        reduction="first",
    ):
        """encode paper from text information and run forward to get sequence and pool output for each entity

        Args:
            title (str, optional): [paper title]. Defaults to ''.
            abstract (str, optional): [paper abstract]. Defaults to ''.
            venue (str, optional): [paper venue]. Defaults to ''.
            authors (list, optional): [paper author]. Defaults to [].
            concepts (list, optional): [paper concepts]. Defaults to [].
            affiliations (list, optional): [paper affiliations]. Defaults to [].
            decode_span_type (str, optional): [the span type to decode, choose from 'FOS','VENUE','AFF','AUTHOR']. Defaults to 'FOS'.
            decode_span_length (int, optional): [the length of span to decode]. Defaults to 0.
            max_seq_length (int, optional): [maximum sequence length for the input, the context information will be truncated if the total length exceeds this number]. Defaults to 512.
            mask_propmt_text (str, optional): [the prompt text to add after title and abstract]. Defaults to ''.
            reduction (str, optional): [the way to get pooled_output, choose from 'cls','max','mean']. Defaults to 'cls'.

        Raises:
            Exception: [provided inputs are invalid]

        Returns:
            [dictionary of list of dictionary]: {
                'text': text_item,
                'venue': venue_item,
                'authors': [authors_item, authors_item, ...]
                'concepts': [concepts_item, concepts_item, ...]
                'affiliations': [affiliations_item, affiliations_item, ...]
                }
        """
        (
            input_ids,
            input_masks,
            token_type_ids,
            masked_lm_labels,
            position_ids,
            position_ids_second,
            masked_positions,
            num_spans,
        ) = self.build_inputs(
            title=title,
            abstract=abstract,
            venue=venue,
            authors=authors,
            concepts=concepts,
            affiliations=affiliations,
            decode_span_type=decode_span_type,
            decode_span_length=decode_span_length,
            max_seq_length=max_seq_length,
            mask_propmt_text=mask_propmt_text,
        )

        search = {
            "text": [title + abstract],
            "venue": [venue],
            "authors": authors,
            "concepts": concepts,
            "affiliations": affiliations,
        }

        item = {
            "originalText": "",
            "inputText": "",
            "type": "",
            "tokens": [],
            "token_ids": [],
            "sequence_output": [],
            "pooled_output": [],
        }

        output = {"text": [], "venue": [], "authors": [], "concepts": [], "affiliations": []}

        split_index = {"text": [], "venue": [], "authors": [], "concepts": [], "affiliations": []}

        sequence_output, pooled_output = self.bert.forward(
            input_ids=torch.LongTensor(input_ids).unsqueeze(0),
            token_type_ids=torch.LongTensor(token_type_ids).unsqueeze(0),
            attention_mask=torch.LongTensor(input_masks).unsqueeze(0),
            output_all_encoded_layers=False,
            checkpoint_activations=False,
            position_ids=torch.LongTensor(position_ids).unsqueeze(0),
            position_ids_second=torch.LongTensor(position_ids_second).unsqueeze(0),
        )

        entities = {0: "text", 2: "venue", 1: "authors", 4: "concepts", 3: "affiliations"}
        for num, name in entities.items():
            if num in token_type_ids:
                start_index = position_ids[token_type_ids.index(num)]
                split_index[name].append(position_ids.index(start_index) - 1)
                for i in range(0, len(search[name])):
                    split_index[name].append(
                        len(position_ids) - 1 - list(reversed(position_ids)).index(start_index + i)
                    )
                    item = item.copy()
                    item["type"] = name.upper()
                    item["originalText"] = search[name][i]
                    item["token_ids"] = input_ids[split_index[name][i] + 1 : split_index[name][i + 1] + 1]
                    item["tokens"] = self._convert_ids_to_tokens(item["token_ids"])
                    item["inputText"] = self._convert_token_ids_to_text(item["token_ids"])
                    item["sequence_output"] = sequence_output[
                        :, split_index[name][i] + 1 : split_index[name][i + 1] + 1, :
                    ].squeeze(0)
                    if reduction == "mean":
                        item["pooled_output"] = item["sequence_output"].mean(dim=0, keepdim=False)
                    elif reduction == "max":
                        item["pooled_output"], _ = item["sequence_output"].max(dim=0)
                    else:
                        item["pooled_output"] = pooled_output
                    output[name].append(item)

        return output

    def calculate_span_prob(
        self,
        title="",
        abstract="",
        venue="",
        authors=[],
        concepts=[],
        affiliations=[],
        decode_span_type="FOS",
        decode_span="",
        force_forward=False,
        max_seq_length=512,
        mask_propmt_text="",
        device=None,
        debug=False,
    ):
        """calculate span probability by greedy algorithm

        Args:
            title (str, optional): [paper title]. Defaults to ''.
            abstract (str, optional): [paper abstract]. Defaults to ''.
            venue (str, optional): [paper venue]. Defaults to ''.
            authors (list, optional): [paper author]. Defaults to [].
            concepts (list, optional): [paper concepts]. Defaults to [].
            affiliations (list, optional): [paper affiliations]. Defaults to [].
            decode_span_type (str, optional): [the span type to decode, choose from 'FOS','VENUE','AFF','AUTHOR']. Defaults to 'FOS'.
            decode_span_length (int, optional): [the length of span to decode]. Defaults to 0.
            force_forward (bool, optional): [if the decoding order is from left to right]. Defaults to False.
            max_seq_length (int, optional): [maximum sequence length for the input, the context information will be truncated if the total length exceeds this number]. Defaults to 512.
            mask_propmt_text (str, optional): [the prompt text to add after title and abstract]. Defaults to ''.
            device ([type], optional): [device for the inputs, default to cpu]. Defaults to None.
            debug (bool, optional): [if debug is true, instances will print to console]. Defaults to False.

        Returns:
            [tuple[float, list of floats]]: [the probability of the whole span, and the individual probability for each token in the decoded span]
        """
        decode_span_token_ids = self._convert_text_to_token_ids(decode_span)
        decode_span_length = len(decode_span_token_ids)
        (
            input_ids,
            input_masks,
            token_type_ids,
            masked_lm_labels,
            position_ids,
            position_ids_second,
            masked_positions,
            num_spans,
        ) = self.build_inputs(
            title=title,
            abstract=abstract,
            venue=venue,
            authors=authors,
            concepts=concepts,
            affiliations=affiliations,
            decode_span_type=decode_span_type,
            decode_span_length=decode_span_length,
            max_seq_length=max_seq_length,
            mask_propmt_text=mask_propmt_text,
        )

        logprobs = 0.0
        logproblist = []

        def tensorize(x):
            return torch.LongTensor(x).unsqueeze(0).to(device or "cpu")

        for i in range(decode_span_length):
            sequence_output, pooled_output = self.bert.forward(
                input_ids=tensorize(input_ids),
                token_type_ids=tensorize(token_type_ids),
                attention_mask=tensorize(input_masks),
                output_all_encoded_layers=False,
                checkpoint_activations=False,
                position_ids=tensorize(position_ids),
                position_ids_second=tensorize(position_ids_second),
            )
            masked_token_indexes = torch.nonzero((tensorize(masked_lm_labels) + 1).view(-1), as_tuple=False).view(-1)
            prediction_scores, _ = self.cls(sequence_output, pooled_output, masked_token_indexes)
            prediction_scores = torch.nn.functional.log_softmax(prediction_scores, dim=1)  # L x Vocab
            token_log_probs = prediction_scores[torch.arange(len(decode_span_token_ids)), decode_span_token_ids]
            if force_forward:
                logprob, pos = token_log_probs[0], 0
            else:
                logprob, pos = token_log_probs.max(dim=0)
            logprobs += logprob.item()
            logproblist.append(logprob.item())
            real_pos = masked_positions[pos]
            target_token = decode_span_token_ids[pos]
            input_ids[real_pos] = decode_span_token_ids[pos]
            masked_lm_labels[real_pos] = -1
            masked_positions.pop(pos)
            decode_span_token_ids.pop(pos)
            if debug:
                self.print_oag_instance(
                    input_ids=input_ids,
                    token_type_ids=token_type_ids,
                    input_masks=input_masks,
                    masked_lm_labels=masked_lm_labels,
                    position_ids=position_ids,
                    position_ids_second=position_ids_second,
                    predictions=torch.topk(prediction_scores, k=5, dim=1).indices.cpu().detach().numpy(),
                )
                input(
                    "logprobs: %.4f, logprob: %.4f, pos: %d, real_pos: %d, token: %s"
                    % (logprobs, logprob, pos.item(), real_pos, self._convert_ids_to_tokens([target_token])[0])
                )

        return np.exp(logprobs), logproblist

    def decode_beamsearch(
        self,
        title="",
        abstract="",
        venue="",
        authors=[],
        concepts=[],
        affiliations=[],
        decode_span_type="",
        decode_span_length=0,
        beam_width=16,
        force_forward=False,
        max_seq_length=512,
        mask_propmt_text="",
        device=None,
        debug=False,
    ):
        """decode span by using beamsearch

        Args:
            title (str, optional): [paper title]. Defaults to ''.
            abstract (str, optional): [paper abstract]. Defaults to ''.
            venue (str, optional): [paper venue]. Defaults to ''.
            authors (list, optional): [paper author]. Defaults to [].
            concepts (list, optional): [paper concepts]. Defaults to [].
            affiliations (list, optional): [paper affiliations]. Defaults to [].
            decode_span_type (str, optional): [the span type to decode, choose from 'FOS','VENUE','AFF','AUTHOR']. Defaults to 'FOS'.
            decode_span_length (int, optional): [the length of span to decode]. Defaults to 0.
            beam_width (int, optional): [beam search width, notice that this function will run one step of beam search in a batch, which should ensure that your gpu (if using) should be able to hold this number of instances]. Defaults to 16.
            force_forward (bool, optional): [if the decoding order is from left to right]. Defaults to False.
            max_seq_length (int, optional): [maximum sequence length for the input, the context information will be truncated if the total length exceeds this number]. Defaults to 512.
            mask_propmt_text (str, optional): [the prompt text to add after title and abstract]. Defaults to ''.
            device ([type], optional): [device for the inputs, default to cpu]. Defaults to None.
            debug (bool, optional): [if debug is true, the beam search progress will be shown]. Defaults to False.

        Returns:
            [list of (string, float)]: [a list of generated spans with their probablities]
        """
        (
            input_ids,
            input_masks,
            token_type_ids,
            masked_lm_labels,
            position_ids,
            position_ids_second,
            masked_positions,
            num_spans,
        ) = self.build_inputs(
            title=title,
            abstract=abstract,
            venue=venue,
            authors=authors,
            concepts=concepts,
            affiliations=affiliations,
            decode_span_type=decode_span_type,
            decode_span_length=decode_span_length,
            max_seq_length=max_seq_length,
            mask_propmt_text=mask_propmt_text,
        )

        q = [(input_ids, masked_lm_labels, masked_positions, 0)]

        def tensorize(x):
            return torch.LongTensor(x).to(device or "cpu")

        for i in range(decode_span_length):
            sequence_output, pooled_output = self.bert.forward(
                input_ids=tensorize([_input_ids for _input_ids, _, _, _ in q]),
                token_type_ids=tensorize([token_type_ids for _ in q]),
                attention_mask=tensorize([input_masks for _ in q]),
                output_all_encoded_layers=False,
                checkpoint_activations=False,
                position_ids=tensorize([position_ids for _ in q]),
                position_ids_second=tensorize([position_ids_second for _ in q]),
            )
            masked_token_indexes = torch.nonzero(
                (tensorize([_masked_lm_labels for _, _masked_lm_labels, _, _ in q]) + 1).view(-1), as_tuple=False
            ).view(-1)
            prediction_scores, _ = self.cls(sequence_output, pooled_output, masked_token_indexes)
            prediction_scores = torch.nn.functional.log_softmax(
                prediction_scores, dim=1
            )  # (len(q) * (range_length - i), VOCAB_SIZE)
            vocab_size = prediction_scores.shape[-1]
            _q = []
            mask_length = decode_span_length - i
            for idx, (_input_ids, _masked_lm_labels, _masked_positions, _last_logprob) in enumerate(q):
                if force_forward:
                    log_probs, indices = torch.topk(prediction_scores[idx * mask_length].view(-1), k=beam_width)
                else:
                    log_probs, indices = torch.topk(
                        prediction_scores[idx * mask_length : (idx + 1) * mask_length].view(-1), k=beam_width
                    )
                for log_prob, index in zip(log_probs.detach().numpy(), indices.detach().numpy()):
                    new_input_ids = _input_ids.copy()
                    new_masked_lm_labels = _masked_lm_labels.copy()
                    new_masked_positions = _masked_positions.copy()
                    fill_id = index % vocab_size
                    rel_fill_pos = index // vocab_size
                    fill_pos = _masked_positions[rel_fill_pos]
                    new_input_ids[fill_pos] = fill_id
                    new_masked_lm_labels[fill_pos] = -1
                    new_masked_positions.pop(rel_fill_pos)
                    _q.append((new_input_ids, new_masked_lm_labels, new_masked_positions, _last_logprob + log_prob))
            _q.sort(key=lambda tup: tup[-1], reverse=True)
            keys = set()
            q.clear()
            for tup in _q:
                key = tuple(tup[0][-decode_span_length:])
                if key not in keys:
                    keys.add(key)
                    q.append(tup)
                    if len(q) >= beam_width:
                        break

            if debug:
                print("beam search, rest length: %d" % (decode_span_length - i))
                for _input_ids, _masked_lm_labels, _masked_positions, _last_logprob in q:
                    print(
                        "  %8.4f, %s"
                        % (
                            _last_logprob,
                            self._convert_token_ids_to_string(_input_ids[-decode_span_length:]),
                        )
                    )

        results = []
        for (_input_ids, _masked_lm_labels, _masked_positions, _logprob) in q:
            generated_entity = self._convert_token_ids_to_text(_input_ids[-decode_span_length:])
            results.append((generated_entity, np.exp(_logprob)))
        return results

    def generate_title(  # noqa C901
        self,
        abstract="",
        authors=[],
        venue="",
        affiliations=[],
        concepts=[],
        num_beams=1,
        no_repeat_ngram_size=3,
        num_return_sequences=1,
        min_length=10,
        max_length=30,
        device=None,
        early_stopping=False,
        debug=False,
    ):
        """generate paper titles given other information

        Args:
            abstract (str, optional): [paper abstract]. Defaults to ''.
            venue (str, optional): [paper venue]. Defaults to ''.
            authors (list, optional): [paper author]. Defaults to [].
            affiliations (list, optional): [paper affiliations]. Defaults to [].
            concepts (list, optional): [paper concepts]. Defaults to [].
            num_beams (int, optional): [beam search width, notice that this function will run one step of beam search in a batch, which should ensure that your gpu (if using) should be able to hold this number of instances]. Defaults to 1.
            no_repeat_ngram_size (int, optional): [n-grams phrases cannot repeat in title]. Defaults to 3.
            num_return_sequences (int, optional): [number of sequences to return]. Defaults to 1.
            min_length (int, optional): [the minimum length of generated title]. Defaults to 10.
            min_length (int, optional): [the maximum length of generated title]. Defaults to 30.
            early_stopping (bool, optional): [terminate generation while target number of generated sequences reach <EOS>]. Defaults to false.
            device ([type], optional): [device for the inputs, default to cpu]. Defaults to None.
            debug (bool, optional): [if debug is true, the beam search progress will be shown]. Defaults to False.

        Returns:
            [list of (string, float)]: [a list of generated titles with their probablities]
        """
        if num_return_sequences > num_beams:
            raise Exception(
                "num_return_sequences(%d) cannot be larger than num_beams(%d)" % (num_return_sequences, num_beams)
            )

        selected_ngrams = {}
        mask_token_id = self.tokenizer.mask_token_id
        eos_token_id = 1
        token_type_id = 0

        (
            input_ids,
            input_masks,
            token_type_ids,
            masked_lm_labels,
            position_ids,
            position_ids_second,
            masked_positions,
            num_spans,
        ) = self.build_inputs(
            title="[CLS] [SEP]",
            abstract=abstract,
            venue=venue,
            authors=authors,
            concepts=concepts,
            affiliations=affiliations,
            decode_span_type="TEXT",
            decode_span_length=0,
            max_seq_length=512,
            mask_propmt_text="",
        )

        context_length = len(input_ids)
        num_spans = 0
        decode_pos = 1
        decode_postion_ids_second = 1
        for i in range(1, context_length):
            if token_type_ids[i] == 0:
                position_ids_second[i] = i + 1

        input_ids.insert(decode_pos, mask_token_id)
        token_type_ids.insert(decode_pos, token_type_id)
        position_ids.insert(decode_pos, num_spans)
        position_ids_second.insert(decode_pos, decode_postion_ids_second)
        masked_lm_labels.insert(decode_pos, self.tokenizer.cls_token_id)

        q = [(input_ids, 0)]
        selected_entities = []

        def tensorize(x):
            return torch.LongTensor(x).to(device or "cpu")

        while True:
            batch_input_ids = tensorize([_input_ids for _input_ids, _ in q])
            batch_token_type_ids = tensorize([token_type_ids for _ in q])

            current_total_length = batch_input_ids.shape[1]
            current_entity_length = current_total_length - context_length

            batch_attention_mask = torch.ones((current_total_length, current_total_length))
            batch_attention_mask[
                decode_pos - current_entity_length + 1 : decode_pos + 1,
                decode_pos - current_entity_length + 1 : decode_pos + 1,
            ] = torch.tril(
                batch_attention_mask[
                    decode_pos - current_entity_length + 1 : decode_pos + 1,
                    decode_pos - current_entity_length + 1 : decode_pos + 1,
                ]
            )
            batch_attention_mask = batch_attention_mask.unsqueeze(0).repeat(len(q), 1, 1).to(device or "cpu")

            batch_position_ids = tensorize([position_ids for _ in q])
            batch_position_ids_second = tensorize([position_ids_second for _ in q])
            batch_masked_lm_labels = tensorize([masked_lm_labels for _ in q])
            sequence_output, pooled_output = self.bert.forward(
                input_ids=batch_input_ids,
                token_type_ids=batch_token_type_ids,
                attention_mask=batch_attention_mask,
                output_all_encoded_layers=False,
                checkpoint_activations=False,
                position_ids=batch_position_ids,
                position_ids_second=batch_position_ids_second,
            )
            masked_token_indexes = torch.nonzero((batch_masked_lm_labels + 1).view(-1)).view(-1)
            prediction_scores, _ = self.cls(sequence_output, pooled_output, masked_token_indexes)
            prediction_scores = torch.nn.functional.log_softmax(prediction_scores, dim=1)
            # surpress existing n-grams
            for idx, (_input_ids, _) in enumerate(q):
                if current_entity_length >= no_repeat_ngram_size:
                    prefix_key = tuple(_input_ids[decode_pos - no_repeat_ngram_size + 1 : decode_pos])
                    for token_id in selected_ngrams.get(prefix_key, set()):
                        prediction_scores[idx, token_id] = -10000
                prefix_key = tuple(_input_ids[decode_pos - current_entity_length : decode_pos])
                if prefix_key in selected_ngrams:
                    for token_id in selected_ngrams.get(prefix_key, set()):
                        prediction_scores[idx, token_id] = -10000
                if current_entity_length <= min_length:
                    prediction_scores[idx, eos_token_id] = -10000
                prediction_scores[idx, _input_ids[decode_pos]] = -10000

            decode_pos += 1
            _q = []
            log_probs, indices = torch.topk(prediction_scores, k=num_beams)
            for idx, (_input_ids, _last_logprob) in enumerate(q):
                for k in range(log_probs.shape[1]):
                    new_input_ids = _input_ids.copy()
                    new_input_ids.insert(decode_pos, indices[idx, k].item())
                    _q.append((new_input_ids, _last_logprob + log_probs[idx, k].item()))

            q = []
            for _input_ids, _last_logprob in _q:
                prefix_key = None
                if current_entity_length >= no_repeat_ngram_size:
                    prefix_key = tuple(_input_ids[decode_pos - no_repeat_ngram_size + 1 : decode_pos])
                    if prefix_key not in selected_ngrams:
                        selected_ngrams[prefix_key] = set()
                    selected_ngrams[prefix_key].add(_input_ids[decode_pos])
                if _input_ids[decode_pos] == eos_token_id:
                    selected_entities.append((_input_ids, _last_logprob))
                else:
                    q.append((_input_ids, _last_logprob))
            q.sort(key=lambda tup: tup[-1], reverse=True)
            selected_entities.sort(key=lambda tup: tup[-1], reverse=True)
            q = q[:num_beams]
            if current_entity_length >= max_length + 2:
                break
            if len(selected_entities) >= num_return_sequences:
                if early_stopping or len(q) == 0 or q[0][-1] <= selected_entities[num_return_sequences - 1][-1]:
                    break

            token_type_ids.insert(decode_pos, token_type_id)
            position_ids.insert(decode_pos, num_spans)
            position_ids_second.insert(decode_pos, decode_postion_ids_second)
            masked_lm_labels[decode_pos - 1] = -1
            masked_lm_labels.insert(decode_pos, self.tokenizer.cls_token_id)

            if debug:
                self.print_oag_instance(
                    input_ids=batch_input_ids[0].cpu().detach().numpy(),
                    token_type_ids=batch_token_type_ids[0].cpu().detach().numpy(),
                    input_masks=batch_attention_mask[0].cpu().detach().numpy(),
                    masked_lm_labels=batch_masked_lm_labels[0].cpu().detach().numpy(),
                    position_ids=batch_position_ids[0].cpu().detach().numpy(),
                    position_ids_second=batch_position_ids_second[0].cpu().detach().numpy(),
                    predictions=torch.topk(prediction_scores, k=5, dim=1).indices.cpu().detach().numpy(),
                )
                input("== Press Enter for next step ==")

        results = []
        for seq, logprob in selected_entities[:num_return_sequences]:
            token_ids = []
            for _id in seq[decode_pos - current_entity_length + 1 : decode_pos]:
                if _id != eos_token_id:
                    token_ids.append(_id)
                else:
                    break
            results.append((self._convert_token_ids_to_text(token_ids), logprob))
        return results
